#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""RS422 function library for MDrive stepper motors by Novanta IMS (former
Schneider Electric) set up in party mode.

Each MDrive motor has to be flashed using "Novanta Motion Terminal" software
with specific parameters and subroutines set. See the manual
"Programming and Software Reference Manual, MCode, MDrive and MForce Products"
by Novanta IMS.

Required flashed motor parameters:

- No check-sum communication: param CK = 0.
- Escape flag set to [Esc]-key: param ES = 1.
- The echo mode - param EM - must be either: 0 (full-duplex), 1 (half-duplex) or
  3 (print queue). Echo mode EM = 2 is not supported and will result in failing
  to auto-detect the MDrive serial port. This library will change the mode to
  EM = 1 (half-duplex) when method `MDrive_Controller.begin()` has been called.
- All motors must be configured in the so-called 'party mode': param PY = 1.
  Each motor must have been given a unique device name as a single alphanumeric
  character, e.g. param DN = "x", DN = "1", etc. This allows the motors to be
  adressed individually over a single shared serial port.
- User subroutine with label 'F1' should be present within each motor, effecting
  an 'init interface' routine where the MDrive motor should be scripted to stop
  any motion. Advertently, any pending errors of the MDrive get reset by default
  when any subroutine gets executed as per MDrive design.

  Minimal needed MCode::

    LB F1             '--- Init interface
        SL 0          'Stop any movement
        H             'Wait for movement to finish
        BR M0         'Goto: Main loop

- User subroutine with label 'F2' should be present within each motor, effecting
  a 'home' routine where the MDrive motor should be scripted to perform a homing
  operation and have the internal step counter be reset to 0.

  Minimal needed MCode::

    LB F2             '--- Home
        SL 0            'Stop any movement
        H               'Wait for movement to finish
        HM 1            'Home using method 1 (adjust method to your needs)
        H               'Wait for movement to finish
        R1 = P & 1023
        R2 = 1024 - R1
        MR R2           'Optionally: Move tiny amount away from home again
        H               'Wait for movement to finish
        P = 0           'Redefine position to == 0
        BR M0           'Goto: Main loop

- User variable with label 'C0' should be present indicating the [steps/mm] or
  [steps/rev] calibration constant of the stage the MDrive motor is connected
  to.
  NOTE: This library only supports integer values for 'C0'. Fractional values
  are not supported.
- User variable with label 'CT' should be present indicating the movement type
  (0: linear, 1: angular) of the stage the MDrive motor is connected to.

Physical wiring:

- Because we require each motor to operate in party-mode, the physical serial
  wiring looks as follows. Each TX-, TX+, RX- and RX+ wire per motor can be tied
  together, resulting in them sharing the same serial port. A single
  RS422-to-USB adapter can then be used to control all motors via this library.
  The Novanta manual (Hardware Manual MDrive Motion Control, chapter 'Multi-drop
  communication connection`) recommends to have only one ground connection to
  the first motor and not to the other motors to prevent ground loops.
  Termination resistors are necessary when the serial cable exceed 4.5 meters.

NOTE: Regarding the obsolete MDrive stepper control box designed by Rindert
Nauta of the TCO department of the University of Twente. It is a custom-build
hardware interface with a user-control panel that interfaces with multiple
MDrive stepper motors. It allows for stand-alone control of each motor via push
buttons, rotary knobs and a LCD screen. The controller is based on a PIC
microcontroller with source-code programmed by Rindert Nauta. The source code is
unknown, but the API is mostly known. The controller allows for direct
pass-through via serial RS232 and USB to a host PC, allowing motor control via
scripting, bypassing the user-control panel. Holding down the 'Home' button
during power on will set the RS232 port in direct pass-through mode to each
attached motor, allowing them to be flashed via a PC.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=missing-function-docstring, too-many-lines

import sys
import time
from enum import Enum
from typing import Union, Tuple, Dict

import numpy as np
import serial

from dvg_debug_functions import dprint, ANSI
from dvg_devices.BaseDevice import SerialDevice

# Print extra debugging info to the terminal? Use only for troubleshooting.
DEBUG = False


class Movement_type(Enum):
    LINEAR = 0
    ANGULAR = 1


def print_warning(msg: str):
    print(f"{ANSI.RED}{msg}{ANSI.WHITE}")


def shortest_path_on_step_circle(
    from_step: int,
    to_step: int,
    steps_per_rev: int,
) -> int:
    """Return the relative integer step distance on a circle consisting of
    `steps_per_rev` steps per full revolution that results in the shortest path
    starting from position `from_step` to ending position `to_step`.

    Returns:
        dist (`int`):
            The relative integer step distance ranging from
            `(-steps_per_rev // 2, steps_per_rev // 2]` where a positive sign
            indicates CCW rotation and a negative sign CW rotation.

    NOTE: No floating point rounding errors. All integer maths.
    """
    # Normalize each to [0, steps_per_rev)
    from_step %= steps_per_rev
    to_step %= steps_per_rev

    dist = (to_step - from_step) % steps_per_rev  # CCW
    dist_CW = steps_per_rev - dist

    if dist > dist_CW:
        dist = -dist_CW

    return dist


def shortest_path_on_unit_circle(from_rev: float, to_rev: float) -> float:
    """Return the relative distance on a unit circumference circle that results
    in the shortest path starting from position `from_rev` to ending position
    `to_rev` expressed in unit revolutions.

    Returns:
        dist (`float`):
            The relative distance ranging from (-0.5, 0.5] on a unit circle
            where a positive sign indicates CCW rotation and a negative sign CW
            rotation.

    NOTE: Floating point rounding errors are unavoidable, e.g.,
    ``shortest_path_on_unit_circle(0.1, 0.3) = 0.19999999999999998``
    """
    # Normalize each to [0, 1)
    from_rev %= 1
    to_rev %= 1

    dist = (to_rev - from_rev) % 1  # CCW
    dist_CW = 1 - dist

    if dist > dist_CW:
        dist = -dist_CW

    return dist


# ------------------------------------------------------------------------------
#   MDrive_Controller
# ------------------------------------------------------------------------------


class MDrive_Controller(SerialDevice):
    """Software controller class to interface with multiple MDrive motors set up
    in party mode over a single serial port.

    Main methods:
    - auto_connect()
    - begin()
    - STOP()
    - RESET()
    - close()

    Main members:
    - ser
    - motors

    Inherits from `dvg_devices.BaseDevice.SerialDevice`. See there for more info
    on other arguments, methods and members.
    """

    def __init__(
        self,
        name="MDrive",
        long_name="MDrive controller",
    ):
        super().__init__(name=name, long_name=long_name)

        # Default serial settings: 9600 8-N-1
        self.serial_settings = {
            "baudrate": 9600,
            "timeout": 0.1,
            "write_timeout": 0.1,
            "parity": "N",
        }
        self.set_read_termination("\n")
        self.set_write_termination("\n")

        self.set_ID_validation_query(
            ID_validation_query=self.ID_validation_query,
            valid_ID_broad="MDrive",
            valid_ID_specific=None,
        )

        # List of all detected motors
        self.motors: list[MDrive_Motor] = []

    # --------------------------------------------------------------------------
    #   flush_serial_in
    # --------------------------------------------------------------------------

    def flush_serial_in(self):
        """Silently flush the serial in buffer of the OS."""
        try:
            # Ensure the MDrive has had enough time to write any pending bits
            # into the serial buffer of the OS
            time.sleep(0.1)

            reply = self.ser.read_all()
            if DEBUG:
                dprint(f"flush: {reply}", ANSI.CYAN)
        except:  # pylint: disable=bare-except
            pass

    # --------------------------------------------------------------------------
    #   ID_validation_query
    # --------------------------------------------------------------------------

    def ID_validation_query(
        self,
    ) -> Tuple[Union[str, bytes, None], Union[str, bytes, None]]:
        r"""Sends the escape character "\x1b" which stops all current motion of
        the MDrive motors to which we expect any of the following replies::

          b"#\r\n>"   Reply when in full-duplex (EM = 0) without queued error
          b"#\r\n?"   Reply when in full-duplex (EM = 0) with queued error
          b"\r\n"     Reply when in half-duplex (EM = 1)
        """
        # `self.query()` reads until the '\n' character is encountered
        _, reply = self.query("\x1b", returns_ascii=False)  # Sending [Esc]
        if DEBUG:
            dprint(f"\nID   : {reply}", ANSI.CYAN)

        if isinstance(reply, bytes):
            if reply[:2] == b"\r\n" or reply[:3] == b"#\r\n":
                # We have to flush the serial in buffer of the OS, because there
                # might still be pending characters to be read. Namely '>' or
                # '?', which we will ignore. Sometimes the remaining bits get
                # garbled up, preventing decoding into ASCII characters. This
                # could happen when multiple motors respond simultaneously on
                # the same serial port, garbling up the ASCII bit stream.
                self.flush_serial_in()

                # ID validation successful
                return "MDrive", None

        # ID validation failed
        return "", None

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(
        self,
        device_names_to_scan: Union[str, None] = None,
    ):
        """Scan for connected motors, set them up and store them in member
        `MDrive_Controller.motors` as a list of `MDrive_Motor` instances.

        This method should be called directly after having established a
        serial connection to the motors, e.g. after method
        `MDrive_Controller.auto_connect()`.

        Args:
            device_names_to_scan (`str`, optional):
                List of alphanumeric characters that should be scanned over to
                detect motors by their device name. When omitted, will scan over
                all 62 alphanumeric characters, taking ~7 seconds. To speed up
                the scan, one can limit the list to specific device names if
                they are known a-priori, e.g., "xyza".

                Default: All alphanumeric characters [0-9][a-z][A-Z].
        """

        # Step 1: Scan for attached motors
        """
        Expected replies when motor is attached:
          b"[DN]\r\n>"  Reply when in full-duplex (EM = 0) without queued error
          b"[DN]\r\n?"  Reply when in full-duplex (EM = 0) with queued error
          b"\r\n"       Reply when in half-duplex (EM = 1)
        Expected replies when motor not found:
          b""           Reply for both half- and full-duplex
          or a serial read time-out.
        """

        if device_names_to_scan is None:
            device_names_to_scan = (
                "0123456789"
                "abcdefghijklmnopqrstuvwxyz"
                "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            )
            print("Scanning for all motors. Takes ~7 seconds...")
        else:
            print(f"Scanning for motors in list '{device_names_to_scan}'...")

        for scan_DN in device_names_to_scan:
            try:
                success, reply = self.query(
                    scan_DN,
                    raises_on_timeout=True,
                    returns_ascii=False,
                )
                if DEBUG:
                    dprint(f"scan : {reply}", ANSI.CYAN)

            except serial.SerialException:
                # Due to a serial read time-out, or an empty bytes b"" reply.
                # In both cases: There is no motor attached.
                pass

            else:
                if success:
                    # Flush possibly remaining '>', '?' chars left in the buffer
                    # when the MDrive is in full-duplex mode (EM = 0).
                    self.flush_serial_in()

                    print(
                        f"{ANSI.YELLOW}  - DETECTED MOTOR '{scan_DN}'"
                        f"{ANSI.WHITE}"
                    )

                    self.motors.append(
                        MDrive_Motor(
                            controller=self,
                            device_name=scan_DN,
                        )
                    )
                    self.motors[-1].begin()

                    print("")

        # Report
        print("Done scanning for motors. ", end="")
        N_motors = len(self.motors)
        if N_motors == 0:
            print_warning("NO MOTORS DETECTED.\n")
        else:
            pretty_list = ", ".join(
                [f"'{motor.device_name}'" for motor in self.motors]
            )
            print(
                f"Detected {N_motors} motor{'s' if N_motors > 1 else ''}: "
                f"{pretty_list}.\n"
            )

    # --------------------------------------------------------------------------
    #   STOP
    # --------------------------------------------------------------------------

    def STOP(self):
        """Emergency stop any motion of all MDrive motors.

        NOTE: To resume normal operation after an emergency stop it is necessary
        to call method `RESET()`.
        """
        if DEBUG:
            dprint("STOP! Sending [Esc] to all MDrive motors.", ANSI.CYAN)

        self.ser.write(b"\x1b")  # [Esc]

        # We flush because multiple MDrives can respond simultaneously, garbling
        # up the serial bit stream. No point in trying to read the reply.
        self.flush_serial_in()

    # --------------------------------------------------------------------------
    #   RESET
    # --------------------------------------------------------------------------

    def RESET(self):
        """Reset all MDrive motors. It allows for normal operation to resume
        after an emergency stop via method `STOP()` was issued. Reset will also
        effect a controlled stop of any motion of all motors.

        NOTE: User subroutine 'F1' will be executed by all MDrive motors.
        """
        if DEBUG:
            dprint("RESET! Sending 'EX F1' to all MDrive motors.", ANSI.CYAN)

        self.ser.write(b"*EX F1\n")

        # We flush because multiple MDrives can respond simultaneously, garbling
        # up the serial bit stream. No point in trying to read the reply.
        self.flush_serial_in()

    # --------------------------------------------------------------------------
    #   close
    # --------------------------------------------------------------------------

    def close(self, ignore_exceptions=False):
        """Controlled-stop any motion of all motors and close the serial port."""
        if self.is_alive:
            self.RESET()
        super().close(ignore_exceptions)

    # --------------------------------------------------------------------------
    #   query_half_duplex
    # --------------------------------------------------------------------------

    def query_half_duplex(self, msg: str) -> Tuple[bool, str]:
        r"""Send a message over the MDrive serial port and subsequently read the
        reply.

        Wrapped version of method `dvg_devices.BaseDevice.SerialDevice.query()`
        that requires all MDrive motors to be set up in half-duplex mode
        (EM = 1), which should be the case when a call to method `begin()` has
        been performed.

        This method will check for a proper reply coming back from the MDrive
        motor, i.e. the raw reply should end with bytes b'\r\n'. When serial
        communication has failed or the proper end bytes are not received
        a warning message will be printed to the terminal, but the program will
        continue on.

        Args:
            msg (`str`):
                ASCII string to be sent over the MDrive serial port.

        Returns:
            `tuple`:
                success (`bool`):
                    True if successful, False otherwise.

                reply (`str`):
                    Reply received from the device as an ASCII string.
        """
        _, reply = self.query(msg, returns_ascii=False)

        if DEBUG:
            dprint(f"msg: {msg}", ANSI.GREEN)
            dprint(f"└> {reply}", ANSI.YELLOW)

        if isinstance(reply, bytes) and reply[-2:] == b"\r\n":
            try:
                reply = reply[:-2].decode()
                return True, reply

            except UnicodeDecodeError:
                dprint("MDRIVE COMMUNICATION ERROR", ANSI.RED)
                dprint(f"  Trying to send: {msg}")
                dprint(f"  Received reply: {reply}")
                return False, ""

        dprint("MDRIVE COMMUNICATION ERROR", ANSI.RED)
        dprint(f"  Trying to send: {msg}")
        dprint(f"  Received reply: {reply}")
        return False, ""


# ------------------------------------------------------------------------------
#   MDrive_Motor
# ------------------------------------------------------------------------------


class MDrive_Motor:
    """Class to manage communication with a single MDrive motor set up in party
    mode.

    Main members:
    - config
    - state

    Args:
        controller (`MDrive_Controller`):
            Parent controller in charge of the serial port to which the MDrive
            motor is connected to.

        device_name (`str`):
            Device name of this particular MDrive motor (param DN).
    """

    class Config:
        """Container for the MDrive motor configuration parameters."""

        part_number: str = ""
        """Param PN"""
        serial_number: str = ""
        """Param SN"""
        firmware_version: str = ""
        """Param VR"""

        IO_S1: str = ""
        """Setup IO Point 1"""
        IO_S2: str = ""
        """Setup IO Point 2"""
        IO_S3: str = ""
        """Setup IO Point 3"""
        IO_S4: str = ""
        """Setup IO Point 4"""

        user_variables: Dict[str, int] = {}
        """Param UV: Dictionary of user variables"""
        user_subroutines: Dict[str, int] = {}
        """Param UV: Dictionary of user subroutines"""

        motion_A: Union[float, int] = np.nan
        """Acceleration [steps/sec^2]"""
        motion_D: Union[float, int] = np.nan
        """Deceleration [steps/sec^2]"""
        motion_HC: Union[float, int] = np.nan
        """Hold current [0-100 %]"""
        motion_HT: Union[float, int] = np.nan
        """Hold current delay time [msec]"""
        motion_LM: Union[float, int] = np.nan
        """Limit stop modes [1-6]"""
        motion_MS: Union[float, int] = np.nan
        """Microstep resolution select [microsteps per full motor step]"""
        motion_MT: Union[float, int] = np.nan
        """Motor settling delay time [msec]"""
        motion_RC: Union[float, int] = np.nan
        """Run current [0-100 %]"""
        motion_VI: Union[float, int] = np.nan
        """Initial velocity [steps/sec]"""
        motion_VM: Union[float, int] = np.nan
        """Maximum velocity [steps/sec]"""

        movement_type: Movement_type = Movement_type.LINEAR
        """Read from MDrive user variable 'CT'. 0: linear, 1: angular."""
        calibration_constant: Union[float, int] = np.nan
        """Read from MDrive user variable 'C0'. The step calibration constant in
        units of [steps/mm] or [steps/rev], depending on the movement type
        parameter 'CT'."""

    class State:
        """Container for the MDrive motor measurement variables."""

        position: Union[float, int] = np.nan
        """Param P: Read position [steps]"""
        velocity: Union[float, int] = np.nan
        """Param V: Read velocity [steps/sec]"""
        is_moving: bool = False
        """Param MV: Moving flag [True/False]"""
        is_velocity_changing: bool = False
        """Param VC: Velocity changing flag [True/False]"""

        has_error: bool = False
        """Param EF: Error flag [True/False]"""
        error: int = 0
        """Param ER: Error number"""

        is_home_known: bool = False
        """Does not exist as a build-in MDrive motor parameter that we can
        query, but is supplied here as a means to keep (loosely) track of
        whether the `position` parameter is to be trusted or not. The end-user
        is free to set and get the `is_home_known` parameter by themselves for
        whatever reason that is not captured by the following automation:

        This flag only gets automatically set to True whenever method `home()`
        has ran. Also, any reported error by the MDrive motor via a call to
        method `query_errors` will set this flag automatically to False."""

    def __init__(self, controller: MDrive_Controller, device_name: str):
        self.controller = controller
        self.device_name = device_name  # Param DN
        self.config = self.Config()
        self.state = self.State()

        # Placeholder for a future `QtCore.QMutex()` instance to allow for
        # proper multithreading. Will be assigned in `MDrive_stepper_qdev.py`.
        self.mutex: Union[object, None] = None

    # --------------------------------------------------------------------------
    #   query
    # --------------------------------------------------------------------------

    def query(self, msg: str) -> Tuple[bool, str]:
        r"""Send a message to this particular MDrive motor and subsequently read
        the reply.

        NOTE: The message string will automatically be prepended with the device
        name of this particular motor, which is assumed to be configured in
        party mode (PY = 1) and half-duplex mode (EM = 1). This is the case when
        `MDrive_Controller.begin()` has ran successfully.

        Args:
            msg (`str`):
                ASCII string to be sent to the MDrive motor.

        Returns:
            `tuple`:
                success (`bool`):
                    True if successful, False otherwise.

                reply (`str`):
                    Reply received from the device as an ASCII string.
        """
        return self.controller.query_half_duplex(f"{self.device_name}{msg}")

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(self):
        """Set up the MDrive motor for operation.

        The following will take place:
        1) Set the echo mode to half-duplex (EM = 1).
        2) Stop all motion and reset any errors by calling subroutine 'F1'.
        3) Retrieve the configuration and initial measurement parameters of each
        motor.

        This method should be called directly after having established a
        serial connection to the motors, e.g. after method
        `MDrive_Controller.auto_connect()`.
        """
        # Set the echo mode to half-duplex. We don't care about the reply.
        self.query("em 1")

        # Stop any motion and reset errors. Calling this subroutine 'F1' is
        # crucial, because the MDrive motor has a strange quirk as follows.
        #
        # In detail: The [Esc]-character we sent to auto-detect the controller
        # seems to prepend any future query replies with '\r\n', i.e. instead
        # of replying to '1pr p' with '0\r\n' the MDrive motor replies with
        # '\r\n0\r\n'. The latter messes up our query methodology and a call to
        # \any\ subroutine seems to fix the issue. We use subroutine 'F1' here.
        self.init_interface()

        self.ensure_existence_user_variables_C0_and_CT()

        self.query_config()
        self.print_config()

        if "F1" in self.config.user_subroutines:
            print("    Found user subroutine F1: 'Init interface'")
        else:
            print_warning(
                "CRITICAL: User subroutine F1 ('Init interface') was not found."
                " Exiting."
            )
            sys.exit(0)

        if "F2" in self.config.user_subroutines:
            print("    Found user subroutine F2: 'Home'")
        else:
            print_warning(
                "CRITICAL: User subroutine F2 ('Home') was not found. Exiting."
            )
            sys.exit(0)

        self.print_motion_config()

    # --------------------------------------------------------------------------
    #   get_base_unit
    # --------------------------------------------------------------------------

    def get_base_unit(self) -> str:
        """Return a string annotating the base unit derived from MDrive user
        variable 'CT' a follows:
            "mm"  when CT = 0 (linear),
            "rev" when CT = 1 (angular).
        """
        if self.config.movement_type == Movement_type.LINEAR:
            return "mm"

        return "rev"

    # --------------------------------------------------------------------------
    #   ensure_existence_user_variables_C0_and_CT
    # --------------------------------------------------------------------------

    def ensure_existence_user_variables_C0_and_CT(self):
        """Will define the user variables C0 and CT with default values when
        they don't exist in the MDrive flash memory. Unfortunately, they can
        only be declared as local variables which don't persist a program
        restart, unlike global variables."""
        # Default values when not found
        CT = Movement_type.LINEAR
        C0 = 1

        _success, reply = self.query("pr ct")
        try:
            CT = Movement_type(int(reply))
        except ValueError:
            CT_exists = False
        except Exception as e:
            raise e
        else:
            CT_exists = True

        _success, reply = self.query("pr c0")
        try:
            C0 = int(reply)
            C0 = max(1, C0)
        except ValueError:
            C0_exists = False
        except Exception as e:
            raise e
        else:
            C0_exists = True

        self.config.movement_type = CT
        self.config.calibration_constant = C0

        if not CT_exists:
            print_warning(
                "WARNING:\n"
                "  Global user variable CT as movement type (0: linear, 1: "
                "angular) was not found.\n"
                f"  Defaulting to {CT.name}."
            )
            self.query(f"VA CT = {CT.value:d}")
            # In case "CT" was already defined, but out-of-enum-range
            self.query(f"CT {CT.value:d}")

        if not C0_exists:
            print_warning(
                "WARNING:\n"
                "  Global user variable C0 as step calibration constant was "
                "not found.\n"
                f"  Defaulting to 1 [steps/{self.get_base_unit()}].",
            )
            self.query(f"VA C0 = {C0:d}")

    # --------------------------------------------------------------------------
    #   query_config
    # --------------------------------------------------------------------------

    def query_config(self):
        """Query the configuration parameters of the MDrive motor and store
        these inside member `MDrive_Motor.config`.
        """
        # Part number
        success, reply = self.query("pr pn")
        if success:
            self.config.part_number = reply

        # Serial number
        success, reply = self.query("pr sn")
        if success:
            self.config.serial_number = reply

        # Firmware version
        success, reply = self.query("pr vr")
        if success:
            self.config.firmware_version = reply

        # User variables
        # This query is a special case, because the MDrive motor will return
        # all defined user variables in one go, each delimited by '\r\n'
        # characters. Hence, we send a single query for requesting all user
        # variables and then have to empty out the serial in buffer, line for
        # line, until we find the last empty '\r\n' line.
        lines: list[str] = []  # E.g. lines = ("V1 = G 512000", "SU = 100")
        success, reply = self.query("pr uv")
        while isinstance(reply, str) and reply != "":
            lines.append(reply)
            reply = self.controller.ser.read_until(b"\n")

            if DEBUG:
                dprint(f"   {reply}", ANSI.YELLOW)

            try:
                reply = reply.decode().strip()
            except UnicodeDecodeError:
                dprint("MDRIVE COMMUNICATION ERROR", ANSI.RED)
                dprint(f"  `query_config()` failed to decode reply: {reply}")
                reply = ""

        if success:
            # Parse each line into a dict pair: name & int value
            dict_vars: dict[str, int] = {}
            dict_subr: dict[str, int] = {}
            for line in lines:
                parts = line.split("=")
                dict_key = parts[0].strip()
                dict_val = parts[1].strip()
                if dict_val[0] in ("G", "L"):  # G: global var, L: local var
                    dict_vars[dict_key] = int(dict_val[1:].strip())
                else:
                    dict_subr[dict_key] = int(dict_val)

            self.config.user_variables = dict_vars
            self.config.user_subroutines = dict_subr

            # Calibration variables
            CT = dict_vars.get("CT")
            CT = Movement_type(CT) if CT is not None else Movement_type.LINEAR
            C0 = dict_vars.get("C0")
            C0 = C0 if C0 is not None else np.nan
            self.config.movement_type = CT
            self.config.calibration_constant = C0

        # Motion variables
        success, reply = self.query('pr A,"_"D,"_"HC,"_"HT')
        if success:
            parts = reply.split("_")
            self.config.motion_A = int(parts[0].strip())
            self.config.motion_D = int(parts[1].strip())
            self.config.motion_HC = int(parts[2].strip())
            self.config.motion_HT = int(parts[3].strip())

        success, reply = self.query('pr LM,"_"MS,"_"MT,"_"RC')
        if success:
            parts = reply.split("_")
            self.config.motion_LM = int(parts[0].strip())
            self.config.motion_MS = int(parts[1].strip())
            self.config.motion_MT = int(parts[2].strip())
            self.config.motion_RC = int(parts[3].strip())

        success, reply = self.query('pr VI,"_"VM')
        if success:
            parts = reply.split("_")
            self.config.motion_VI = int(parts[0].strip())
            self.config.motion_VM = int(parts[1].strip())

        # IO variables
        success, reply = self.query('pr S1,"_"S2,"_"S3,"_"S4')
        if success:
            parts = reply.split("_")
            self.config.IO_S1 = parts[0].strip()
            self.config.IO_S2 = parts[1].strip()
            self.config.IO_S3 = parts[2].strip()
            self.config.IO_S4 = parts[3].strip()

    # --------------------------------------------------------------------------
    #   print_config
    # --------------------------------------------------------------------------

    def print_config(self):
        """Print the configuration parameters of the MDrive motor to the
        terminal."""
        config = self.config
        print(f"    Part no.  | {config.part_number}")
        print(f"    Serial    | {config.serial_number}")
        print(f"    Firmware  | {config.firmware_version}")

        print("    IO        | ", end="")
        print(f"S1 = {config.IO_S1}")
        print(f"{'':15} S2 = {config.IO_S2}")
        print(f"{'':15} S3 = {config.IO_S3}")
        print(f"{'':15} S4 = {config.IO_S4}")

        print("    User subr | ", end="")
        if len(config.user_subroutines) == 0:
            print("[empty]")
        else:
            print(", ".join(config.user_subroutines.keys()))

        print("    User vars | ", end="")
        if len(config.user_variables) == 0:
            print("[empty]")
        else:
            for idx, (key, val) in enumerate(config.user_variables.items()):
                if idx > 0:
                    print(" " * 16, end="")
                print(f"{key} = {val}")

    # --------------------------------------------------------------------------
    #   print_motion_config
    # --------------------------------------------------------------------------

    def print_motion_config(self):
        """Print the motion configuration parameters of the MDrive motor to the
        terminal."""
        # Shorthands
        config = self.config
        C0 = config.calibration_constant
        base_unit = self.get_base_unit()

        print("    Motion    | ")
        print(
            f"{'':5} Movement type  CT = {config.movement_type.value} "
            f"({config.movement_type.name})"
        )
        print(f"{'':5} Calibration    C0 = {C0:<10} [steps/{base_unit}]")
        print(
            f"{'':5} Acceleration   A  = {config.motion_A:<10} [steps/sec^2] "
            f"= {config.motion_A / C0:<9.4f} [{base_unit}/sec^2]"
        )
        print(
            f"{'':5} Deceleration   D  = {config.motion_D:<10} [steps/sec^2] "
            f"= {config.motion_D / C0:<9.4f} [{base_unit}/sec^2]"
        )
        print(
            f"{'':5} Initial veloc. VI = {config.motion_VI:<10} [steps/sec]   "
            f"= {config.motion_VI / C0:<9.4f} [{base_unit}/sec]"
        )
        print(
            f"{'':5} Maximum veloc. VM = {config.motion_VM:<10} [steps/sec]   "
            f"= {config.motion_VM / C0:<9.4f} [{base_unit}/sec]"
        )
        print(f"{'':5} Microsteps     MS = {config.motion_MS:<10} [microsteps]")
        print(f"{'':5} Limit stop     LM = {config.motion_LM:<10} [mode 1-6]")
        print(f"{'':5} Run current    RC = {config.motion_RC:<10} [0-100 %]")
        print(f"{'':5} Hold current   HC = {config.motion_HC:<10} [0-100 %]")
        print(f"{'':5} Hold delay     HT = {config.motion_HT:<10} [msec]")
        print(f"{'':5} Settling delay MT = {config.motion_MT:<10} [msec]")

    # --------------------------------------------------------------------------
    #   query_errors
    # --------------------------------------------------------------------------

    def query_errors(self) -> bool:
        """Query the error parameters of the MDrive motor and store these inside
        member `MDrive_Motor.state`.

        Updates:
        - state.has_error
        - state.error
        - state.is_home_known (Automatically set to False whenever any error is
          reported.)

        Returns ('bool'):
            True if the command was successfully send to the motor, False
            otherwise.

        Query takes ~0.024 s @ 9600 baud.
        """
        success, reply = self.query('pr EF,"_"ER')
        if success:
            parts = reply.split("_")
            if len(parts) != 2:
                dprint("MDRIVE COMMUNICATION ERROR", ANSI.RED)
                dprint(f"  `query_errors()` failed to split reply: {reply}")
                return False

            try:
                self.state.has_error = bool(int(parts[0].strip()))
                self.state.error = int(parts[1].strip())
            except ValueError:
                dprint("MDRIVE COMMUNICATION ERROR", ANSI.RED)
                dprint(f"  `query_errors()` failed to parse reply: {reply}")
                return False

        if self.state.has_error:
            self.state.is_home_known = False

        return success

    # --------------------------------------------------------------------------
    #   query_state
    # --------------------------------------------------------------------------

    def query_state(self) -> bool:
        """Query the measurement parameters of the MDrive motor and store these
        inside member `MDrive_Motor.state`.

        Updates:
        - state.position
        - state.velocity
        - state.is_moving
        - state.is_velocity_changing

        Returns ('bool'):
            True if the command was successfully send to the motor, False
            otherwise.

        Query takes ~0.050 s @   9600 baud with TCO box.
        Query takes ~0.062 s @   9600 baud with direct RS422-USB cable.
        Query takes ~0.016 s @ 115200 baud with direct RS422-USB cable.
        """
        success, reply = self.query('pr P,"_"V,"_"MV,"_"VC')
        if success:
            parts = reply.split("_")
            if len(parts) != 4:
                dprint("MDRIVE COMMUNICATION ERROR", ANSI.RED)
                dprint(f"  `query_state()` failed to split reply: {reply}")
                return False

            try:
                self.state.position = int(parts[0].strip())
                self.state.velocity = int(parts[1].strip())
                self.state.is_moving = bool(int(parts[2].strip()))
                self.state.is_velocity_changing = bool(int(parts[3].strip()))
            except ValueError:
                dprint("MDRIVE COMMUNICATION ERROR", ANSI.RED)
                dprint(f"  `query_state()` failed to parse reply: {reply}")
                return False

        return success

    # --------------------------------------------------------------------------
    #   query_is_moving
    # --------------------------------------------------------------------------

    def query_is_moving(self) -> bool:
        """Query the `is_moving` parameter of the MDrive motor and store this
        inside member `MDrive_Motor.state`.

        Updates:
        - state.is_moving

        Returns ('bool'):
            True if the command was successfully send to the motor, False
            otherwise.

        Query takes ~0.013 s @   9600 baud with TCO box.
        Query takes ~0.016 s @   9600 baud with direct RS422-USB cable.
        Query takes ~0.016 s @ 115200 baud with direct RS422-USB cable.
        """
        success, reply = self.query("pr MV")
        if success:
            try:
                self.state.is_moving = bool(int(reply.strip()))
            except ValueError:
                dprint("MDRIVE COMMUNICATION ERROR", ANSI.RED)
                dprint(f"  `query_is_moving()` failed to parse reply: {reply}")
                return False

        return success

    # --------------------------------------------------------------------------
    #   execute_subroutine
    # --------------------------------------------------------------------------

    def execute_subroutine(self, subroutine_label: str) -> Tuple[bool, str]:
        """Execute a subroutine or program as flashed into the MDrive motor.

        NOTE: Calling a subroutine inadvertently will reset any error that might
        be pending inside the MDrive motor. This is by Novanta's design. Both
        parameters 'EF' and 'ER' will be set to 0.

        Args:
            subroutine_label (`str`):
                Label of the subroutine or program as flashed into the MDrive
                motor to be executed.

        Returns:
            `tuple`:
                success (`bool`):
                    True if successful, False otherwise.

                reply (`str`):
                    Reply received from the device as an ASCII string.
        """
        return self.query(f"ex {subroutine_label}")

    # --------------------------------------------------------------------------
    #   init_interface
    # --------------------------------------------------------------------------

    def init_interface(self) -> bool:
        """Perform an init interface routine by calling user subroutine 'F1'
        which stops any motion and resets any errors. Additionally, the
        measurement and error parameters of the MDrive motor get queried and
        stored inside member `MDrive_Motor.state`.

        Returns ('bool'):
            True if the command was successfully send to the motor, False
            otherwise.
        """
        success, _reply = self.execute_subroutine("F1")
        self.query_state()
        self.query_errors()

        return success

    # --------------------------------------------------------------------------
    #   home
    # --------------------------------------------------------------------------

    def home(self) -> bool:
        """Perform a homing routine by calling user subroutine 'F2'.
        Additionally, the measurement and error parameters of the MDrive motor
        get queried and stored inside member `MDrive_Motor.state`.

        Returns ('bool'):
            True if the command was successfully send to the motor, False
            otherwise.
        """
        success, _reply = self.execute_subroutine("F2")
        self.state.is_home_known = success
        self.query_state()
        self.query_errors()

        return success

    # --------------------------------------------------------------------------
    #   Move commands
    # --------------------------------------------------------------------------

    def _move(
        self,
        x: float,
        relative: bool = True,
        in_units_of_step: bool = True,
    ) -> bool:
        """Base method to send a move command ('MA' or 'MR') to the motor.

        Args:
            x (`float`):
                Relative distance to move away from current position, or
                absolute position to move to.

            relative (`bool`, optional):
                When True, move relative distance `x` away from current
                position. When False, move to absolute position `x`.

                Default: True.

            in_units_of_step (`bool`, optional):
                When True, `x` is given in units of [steps]. When False, `x`
                is given in units of [mm] for linear movement and [rev] for
                angular movement.

                Default: True.

        Returns ('bool'):
            True if the command was successfully send to the motor, False
            otherwise.
        """
        config = self.config

        # Ensure x is in units of [steps] from now on
        if not in_units_of_step:
            x = x * config.calibration_constant

        # Safety check
        if np.isnan(x):
            print_warning(
                "WARNING: _move() tripped because x=nan. Movement got "
                "cancelled."
            )
            return False

        # Calculate movement
        if relative:
            # Relative movement
            distance = int(np.round(x))
            msg = f"MR {distance}"

        else:
            # Absolute movement
            if config.movement_type == Movement_type.LINEAR:
                position = int(np.round(x))
                msg = f"MA {position}"
            else:
                # Special case: Absolute angular movement. Must solve for
                # shortest path, turning it into a relative movement command.

                # Ensure integer math
                if np.isnan(self.state.position):
                    print_warning(
                        "WARNING: _move() tripped because position=nan. "
                        "Movement got cancelled."
                    )
                    return False

                if np.isnan(config.calibration_constant):
                    print_warning(
                        "WARNING: _move() tripped because steps_per_rev=nan. "
                        "Movement got cancelled."
                    )
                    return False

                distance = shortest_path_on_step_circle(
                    from_step=int(self.state.position),
                    to_step=int(np.round(x)),
                    steps_per_rev=int(config.calibration_constant),
                )
                msg = f"MR {distance}"

        # Send movement command to motor
        success, _reply = self.query(msg)

        # Query and update movement flag
        self.query_is_moving()

        return success

    def move_absolute_steps(self, x: float) -> bool:
        return self._move(x, relative=False, in_units_of_step=True)

    def move_relative_steps(self, x: float) -> bool:
        return self._move(x, relative=True, in_units_of_step=True)

    def move_absolute_mm(self, x: float) -> bool:
        if not self.config.movement_type == Movement_type.LINEAR:
            print_warning(
                "WARNING: move_absolute_mm() got called but movement type is "
                "angular."
            )
        return self._move(x, relative=False, in_units_of_step=False)

    def move_relative_mm(self, x: float) -> bool:
        if not self.config.movement_type == Movement_type.LINEAR:
            print_warning(
                "WARNING: move_relative_mm() got called but movement type is "
                "angular."
            )
        return self._move(x, relative=True, in_units_of_step=False)

    def move_absolute_rev(self, x: float) -> bool:
        if not self.config.movement_type == Movement_type.ANGULAR:
            print_warning(
                "WARNING: move_absolute_rev() got called but movement type is "
                "linear."
            )
        return self._move(x, relative=False, in_units_of_step=False)

    def move_relative_rev(self, x: float) -> bool:
        if not self.config.movement_type == Movement_type.ANGULAR:
            print_warning(
                "WARNING: move_relative_rev() got called but movement type is "
                "linear."
            )
        return self._move(x, relative=True, in_units_of_step=False)

    # --------------------------------------------------------------------------
    #   Slew commands
    # --------------------------------------------------------------------------

    def _slew(
        self,
        v: float,
        in_units_of_step: bool = True,
    ) -> bool:
        """Base method to send a slew command ('SL') to the motor.

        Args:
            v (`float`):
                Velocity to reach. Takes the acceleration and deceleration
                parameters ('A' and 'D') into account.

            in_units_of_step (`bool`, optional):
                When True, `v` is given in units of [steps/sec]. When False,
                `v` is given in units of [mm/sec] for linear movement and
                [rev/sec] for angular movement.

                Default: True.

        Returns ('bool'):
            True if the command was successfully send to the motor, False
            otherwise.
        """

        # Ensure v is in units of [steps/sec] from now on
        if not in_units_of_step:
            v = v * self.config.calibration_constant

        # Safety check
        if np.isnan(v):
            print_warning(
                "WARNING: _slew() tripped because v=nan. Movement got "
                "cancelled."
            )
            return False

        # Calculate movement
        velocity = int(np.round(v))
        msg = f"SL {velocity}"

        # Send movement command to motor
        success, _reply = self.query(msg)

        # Query and update movement flag
        self.query_is_moving()

        return success

    def slew_steps_per_sec(self, v: float) -> bool:
        return self._slew(v, in_units_of_step=True)

    def slew_mm_per_sec(self, v: float) -> bool:
        if not self.config.movement_type == Movement_type.LINEAR:
            print_warning(
                "WARNING: slew_mm_per_sec() got called but movement type is "
                "angular."
            )
        return self._slew(v, in_units_of_step=False)

    def slew_rev_per_sec(self, v: float) -> bool:
        if not self.config.movement_type == Movement_type.ANGULAR:
            print_warning(
                "WARNING: slew_rev_per_sec() got called but movement type is "
                "linear."
            )
        return self._slew(v, in_units_of_step=False)

    # ------------------------------------------------------------------------------
    #   controlled_stop
    # ------------------------------------------------------------------------------

    def controlled_stop(self) -> bool:
        """Bring the motor to a controlled stop."""
        return self._slew(0)

    # ------------------------------------------------------------------------------
    #   save_to_NVM
    # ------------------------------------------------------------------------------

    def save_to_NVM(self) -> bool:
        """Save the currently known parameters to the MDrive non-volatile
        memory. Executing this command will reinitialize the motor and stop all
        motion."""
        header = f"Saving parameters to NVM of MDrive '{self.device_name}':"

        if not self.init_interface():
            print(header, end="")
            dprint(" FAILED.", ANSI.RED)
            return False

        # Wait for all motion to stop before saving to NVM as per manual
        while self.state.is_moving:
            self.query_is_moving()

        # The "S" command takes a long time to reply back, triggering a serial
        # communication timeout if we don't extend the current timeout.
        timeout_backup = self.controller.ser.timeout
        self.controller.ser.timeout = 2  # [sec]
        success, _reply = self.query("s")
        self.controller.ser.timeout = timeout_backup

        success &= self.query_errors()
        success &= not self.state.has_error

        print(header, end="")
        if success:
            dprint(" Success.", ANSI.GREEN)
        else:
            dprint(" FAILED.", ANSI.RED)

        return success


# ------------------------------------------------------------------------------
#   tests
# ------------------------------------------------------------------------------


def test_shortest_path_on_step_circle():
    spr = 6400  # steps_per_rev

    # Test case 1: Same starting and ending position
    assert shortest_path_on_step_circle(0, 0, spr) == 0
    assert shortest_path_on_step_circle(spr + 10, 10, spr) == 0

    # Test case 2: Shortest path is CCW rotation
    assert shortest_path_on_step_circle(0, spr // 4, spr) == spr // 4
    assert shortest_path_on_step_circle(spr * 3 // 4, 0, spr) == spr // 4

    # Test case 3: Shortest path is CW rotation
    assert shortest_path_on_step_circle(spr // 4, 0, spr) == -spr // 4
    assert shortest_path_on_step_circle(0, spr * 3 // 4, spr) == -spr // 4

    # Test case 4: Shortest path crosses 0 degrees
    assert (
        shortest_path_on_step_circle(spr * 9 // 10, spr // 10, spr)
        == spr * 2 // 10
    )
    assert (
        shortest_path_on_step_circle(spr // 10, spr * 9 // 10, spr)
        == -spr * 2 // 10
    )

    print("All test cases passed!")


def test_shortest_path_on_unit_circle():
    eps = 1e-16  # Max. acceptable rounding error

    # Test case 1: Same starting and ending position
    assert shortest_path_on_unit_circle(0, 0) < eps
    assert shortest_path_on_unit_circle(1.1, 0.1) < eps

    # Test case 2: Shortest path is CCW rotation
    assert shortest_path_on_unit_circle(0, 0.25) - 0.25 < eps
    assert shortest_path_on_unit_circle(0.75, 0) - 0.25 < eps

    # Test case 3: Shortest path is CW rotation
    assert shortest_path_on_unit_circle(0.25, 0) + 0.25 < eps
    assert shortest_path_on_unit_circle(0, 0.75) + 0.25 < eps

    # Test case 4: Shortest path crosses 0 degrees
    assert shortest_path_on_unit_circle(0.9, 0.1) - 0.2 < eps
    assert shortest_path_on_unit_circle(0.1, 0.9) + 0.2 < eps

    print("All test cases passed!")


# ------------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    # test_shortest_path_on_step_circle()
    # test_shortest_path_on_unit_circle()

    mdrive = MDrive_Controller()
    if not mdrive.auto_connect(
        filepath_last_known_port="config/port_MDrive.txt"
    ):
        sys.exit(0)

    # mdrive.begin()
    mdrive.begin(device_names_to_scan="xyza")

    if input("Proceed with motor movement? [y/N]").lower() == "y":
        for motor in mdrive.motors:
            # Test: Homing
            # ------------
            print(f"Homing '{motor.device_name}'... ", end="")
            sys.stdout.flush()

            motor.home()
            while motor.state.is_moving:
                motor.query_is_moving()
            print("done.")

            # Update full state
            motor.query_state()
            motor.query_errors()

            # Test: Moving
            # ------------
            print(f"Moving '{motor.device_name}'... ", end="")
            sys.stdout.flush()
            count = 1

            # motor.move_absolute_mm(20)
            motor.slew_mm_per_sec(10)
            while motor.state.is_moving:
                count += 1
                if count == 100:
                    # mdrive.STOP()
                    # mdrive.RESET()
                    motor.controlled_stop()
                motor.query_is_moving()
            print("done.")

            # Update full state
            motor.query_state()
            motor.query_errors()

    mdrive.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyQt/PySide module to provide multithreaded communication and periodical data
acquisition for a Keysight N8700 power supply (PSU).

TODO: Right now, a PID controller on the power output is injected into the
device instance as a object member. We should have the PID controller already
be defined at the class level of 'Keysight_N8700_protocol_SCPI.Keysight_N8700`.
We should also have the PID parameters be passed as arguments, instead of
hard-coded here.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=missing-function-docstring, multiple-statements, broad-except

import time

from qtpy import QtCore, QtGui, QtWidgets as QtWid
from qtpy.QtCore import Signal, Slot  # type: ignore
import numpy as np

import dvg_pyqt_controls as controls
from dvg_debug_functions import dprint, print_fancy_traceback as pft
from dvg_pid_controller import PID_Controller

from dvg_qdeviceio import QDeviceIO, DAQ_TRIGGER
from dvg_devices.Keysight_N8700_protocol_SCPI import Keysight_N8700

# Monospace font
FONT_MONOSPACE = QtGui.QFont("Monospace", 12, weight=QtGui.QFont.Weight.Bold)
FONT_MONOSPACE.setStyleHint(QtGui.QFont.StyleHint.TypeWriter)


# Enumeration
class GUI_input_fields:
    [ALL, OVP_level, V_source, I_source, P_source] = range(5)


class Keysight_N8700_qdev(QDeviceIO):
    """Manages multithreaded communication and periodical data acquisition for
    a Keysight N8700 power supply (PSU), referred to as the 'device'.

    In addition, it also provides PyQt/PySide GUI objects for control of the
    device. These can be incorporated into your application.

    All device I/O operations will be offloaded to 'workers', each running in
    a newly created thread.

    (*): See 'dvg_qdeviceio.QDeviceIO()' for details.

    Args:
        dev:
            Reference to a
            'dvg_devices.Keysight_N8700_protocol_SCPI.Keysight_N8700'
            instance.

        debug:
            Show debug info in terminal? Warning: Slow! Do not leave on
            unintentionally.

    Main GUI objects:
        qgrp (PyQt5.QtWidgets.QGroupBox)
    """

    signal_GUI_input_field_update = Signal(int)

    def __init__(
        self,
        dev: Keysight_N8700,
        DAQ_trigger=DAQ_TRIGGER.INTERNAL_TIMER,
        DAQ_interval_ms=200,
        DAQ_timer_type=QtCore.Qt.TimerType.CoarseTimer,
        critical_not_alive_count=1,
        debug=False,
        **kwargs,
    ):
        super().__init__(dev, **kwargs)  # Pass kwargs onto QtCore.QObject()
        self.dev: Keysight_N8700  # Enforce type: removes `_NoDevice()`

        # Add PID controller on the power output
        # DvG, 25-06-2018: Kp=0.5, Ki=2, Kd=0
        # TODO: Remove hard-coded PID values and have them passed as arguments
        # via a to-be-written class method inside `Keysight_N8700_protocol_SCPI.
        # Keysight_N8700`.
        self.dev.PID_power = PID_Controller(Kp=0.5, Ki=2, Kd=0)

        self.create_worker_DAQ(
            DAQ_trigger=DAQ_trigger,
            DAQ_function=self.DAQ_function,
            DAQ_interval_ms=DAQ_interval_ms,
            DAQ_timer_type=DAQ_timer_type,
            critical_not_alive_count=critical_not_alive_count,
            debug=debug,
        )

        self.create_worker_jobs(jobs_function=self.jobs_function, debug=debug)

        self.create_GUI()
        self.signal_DAQ_updated.connect(self.update_GUI)
        self.signal_GUI_input_field_update.connect(self.update_GUI_input_field)

        # Update GUI immediately, instead of waiting for the first refresh
        self.update_GUI()
        self.update_GUI_input_field()

    # --------------------------------------------------------------------------
    #   DAQ_function
    # --------------------------------------------------------------------------

    def DAQ_function(self) -> bool:
        DEBUG_local = False
        if DEBUG_local:
            tick = time.perf_counter()

        # Clear input and output buffers of the device. Seems to resolve
        # intermittent communication time-outs.
        if self.dev.device is not None:
            self.dev.device.clear()
            time.sleep(0.01)

        # Finish all operations at the device first
        if not self.dev.wait_for_OPC():
            return False

        if not self.dev.query_V_meas():
            return False
        if not self.dev.query_I_meas():
            return False

        # --------------------
        #   Heater power PID
        # --------------------
        # PID controllers work best when the process and control variables have
        # a linear relationship.
        # Here:
        #   Process var: V (voltage)
        #   Control var: P (power)
        #   Relation   : P = R / V^2
        #
        # Hence, we transform P into P_star
        #   Control var: P_star = sqrt(P)
        #   Relation   : P_star = sqrt(R) / V
        # When we assume R remains constant (which is not the case as the
        # resistance is a function of the heater temperature, but the dependence
        # is expected to be insignificant in our small temperature range of 20
        # to 100 deg C), we now have linearized the PID feedback relation.
        self.dev.PID_power.set_mode(
            (self.dev.state.ENA_output and self.dev.state.ENA_PID),
            self.dev.state.P_meas,
            self.dev.state.V_source,
        )

        self.dev.PID_power.setpoint = np.sqrt(self.dev.state.P_source)
        if self.dev.PID_power.compute(np.sqrt(self.dev.state.P_meas)):
            # New PID output got computed -> send new voltage to PSU
            if self.dev.PID_power.output < 1:
                # PSU does not regulate well below 1 V, hence clamp to 0
                self.dev.PID_power.output = 0
            if not self.dev.set_V_source(self.dev.PID_power.output):
                return False
            # Wait for the set_V_source operation to finish.
            # Takes ~ 300 ms to complete with wait_for_OPC.
            if not self.dev.wait_for_OPC():
                return False

        if not self.dev.query_ENA_OCP():
            return False
        if not self.dev.query_status_OC():
            return False
        if not self.dev.query_status_QC():
            return False
        if not self.dev.query_ENA_output():
            return False

        # Explicitly force the output state to off when the output got disabled
        # on a hardware level by a triggered protection or fault.
        if self.dev.state.ENA_output & (
            self.dev.state.status_QC_OV
            | self.dev.state.status_QC_OC
            | self.dev.state.status_QC_PF
            | self.dev.state.status_QC_OT
            | self.dev.state.status_QC_INH
        ):
            self.dev.state.ENA_output = False
            self.dev.set_ENA_output(False)

        if DEBUG_local:
            tock = time.perf_counter()
            dprint(f"{self.dev.name}: done in {tock - tick:.3f}")

        # Check if there are errors in the device queue and retrieve all
        # if any and append these to 'dev.state.all_errors'.
        if DEBUG_local:
            dprint(f"{self.dev.name}: query errors")
            tick = time.perf_counter()
        self.dev.query_all_errors_in_queue()
        if DEBUG_local:
            tock = time.perf_counter()
            dprint(f"{self.dev.name}: stb done in {tock - tick:.3f}")

        return True

    # --------------------------------------------------------------------------
    #   jobs_function
    # --------------------------------------------------------------------------

    def jobs_function(self, func, args):
        if func == "signal_GUI_input_field_update":
            # Special instruction
            self.signal_GUI_input_field_update.emit(*args)
        else:
            # Default job processing:
            # Send I/O operation to the device
            try:
                func(*args)
                self.dev.wait_for_OPC()
            except Exception as err:
                pft(err)

    # --------------------------------------------------------------------------
    #   create GUI
    # --------------------------------------------------------------------------

    def create_GUI(self):
        # Measure
        p = {
            "alignment": QtCore.Qt.AlignmentFlag.AlignRight,
            "font": FONT_MONOSPACE,
        }
        self.V_meas = QtWid.QLabel("0.00  V   ", **p)
        self.I_meas = QtWid.QLabel("0.000 A   ", **p)
        self.P_meas = QtWid.QLabel("0.00  W   ", **p)

        # Source
        p = {
            "maximumWidth": 60,
            "alignment": QtCore.Qt.AlignmentFlag.AlignRight,
        }
        self.pbtn_ENA_output = controls.create_Toggle_button("Output OFF")
        self.pbtn_ENA_output.clicked.connect(self.process_pbtn_ENA_output)
        self.V_source = QtWid.QLineEdit("0.00", **p)
        self.I_source = QtWid.QLineEdit("0.000", **p)
        self.P_source = QtWid.QLineEdit("0.00", **p)
        self.V_source.editingFinished.connect(self.send_V_source_from_textbox)
        self.I_source.editingFinished.connect(self.send_I_source_from_textbox)
        self.P_source.editingFinished.connect(self.set_P_source_from_textbox)
        self.pbtn_ENA_PID = controls.create_Toggle_button(
            "OFF",
            minimumHeight=28,
            minimumWidth=60,
        )
        self.pbtn_ENA_PID.clicked.connect(self.process_pbtn_ENA_PID)

        # Protection
        self.OVP_level = QtWid.QLineEdit("0.000", **p)
        self.OVP_level.editingFinished.connect(self.send_OVP_level_from_textbox)
        self.pbtn_ENA_OCP = controls.create_Toggle_button(
            "OFF",
            minimumHeight=28,
            minimumWidth=60,
        )
        self.pbtn_ENA_OCP.clicked.connect(self.process_pbtn_ENA_OCP)

        # Questionable condition status registers
        self.status_QC_OV = controls.create_tiny_error_LED()
        self.status_QC_OC = controls.create_tiny_error_LED()
        self.status_QC_PF = controls.create_tiny_error_LED()
        self.status_QC_OT = controls.create_tiny_error_LED()
        self.status_QC_INH = controls.create_tiny_error_LED()
        self.status_QC_UNR = controls.create_tiny_error_LED()

        # Operation condition status registers
        self.status_OC_WTG = controls.create_tiny_error_LED()
        self.status_OC_CV = controls.create_tiny_error_LED()
        self.status_OC_CC = controls.create_tiny_error_LED()

        # Final elements
        self.errors = QtWid.QLineEdit("")
        self.errors.setStyleSheet(controls.SS_TEXTBOX_ERRORS)
        self.pbtn_ackn_errors = QtWid.QPushButton("Acknowledge errors")
        self.pbtn_ackn_errors.clicked.connect(self.process_pbtn_ackn_errors)
        self.pbtn_reinit = QtWid.QPushButton("Reinitialize")
        self.pbtn_reinit.clicked.connect(self.process_pbtn_reinit)
        self.pbtn_save_defaults = QtWid.QPushButton("Save")
        self.pbtn_save_defaults.clicked.connect(self.process_pbtn_save_defaults)
        self.lbl_update_counter = QtWid.QLabel("0")

        i = 0
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(0)
        # fmt: off
        grid.addWidget(self.V_meas                       , i, 0, 1, 4); i+=1
        grid.addWidget(self.I_meas                       , i, 0, 1, 4); i+=1
        grid.addWidget(self.P_meas                       , i, 0, 1, 4); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 8)             , i, 0)      ; i+=1
        grid.addWidget(self.pbtn_ENA_output              , i, 0, 1, 4); i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)            , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Source:")           , i, 0, 1, 4); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 4)             , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Voltage")           , i, 0, 1, 2)
        grid.addWidget(self.V_source                     , i, 2)
        grid.addWidget(QtWid.QLabel("V",)                , i, 3)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 2)             , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Current")           , i, 0, 1, 2)
        grid.addWidget(self.I_source                     , i, 2)
        grid.addWidget(QtWid.QLabel("A")                 , i, 3)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 2)             , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Power PID")         , i, 0, 1, 2)
        grid.addWidget(self.pbtn_ENA_PID                 , i, 2,
                       QtCore.Qt.AlignmentFlag.AlignLeft)             ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 2)             , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Power")             , i, 0, 1, 2)
        grid.addWidget(self.P_source                     , i, 2)
        grid.addWidget(QtWid.QLabel("W")                 , i, 3)      ; i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)            , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Protection:")       , i, 0, 1, 4); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 4)             , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("OVP")               , i, 0, 1, 2)
        grid.addWidget(self.OVP_level                    , i, 2)
        grid.addWidget(QtWid.QLabel("V")                 , i, 3)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 2)             , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("OCP")               , i, 0, 1, 2)
        grid.addWidget(self.pbtn_ENA_OCP                 , i, 2,
                       QtCore.Qt.AlignmentFlag.AlignLeft)             ; i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)            , i, 0)      ; i+=1
        grid.addWidget(self.status_QC_OV                 , i, 0)
        grid.addWidget(QtWid.QLabel("OV")                , i, 1)
        grid.addWidget(QtWid.QLabel("| over-voltage")    , i, 2, 1, 2); i+=1
        grid.addWidget(self.status_QC_OC                 , i, 0)
        grid.addWidget(QtWid.QLabel("OC")                , i, 1)
        grid.addWidget(QtWid.QLabel("| over-current")    , i, 2, 1, 2); i+=1
        grid.addWidget(self.status_QC_PF                 , i, 0)
        grid.addWidget(QtWid.QLabel("PF")                , i, 1)
        grid.addWidget(QtWid.QLabel("| AC power failure"), i, 2, 1, 2); i+=1
        grid.addWidget(self.status_QC_OT                 , i, 0)
        grid.addWidget(QtWid.QLabel("OT")                , i, 1)
        grid.addWidget(QtWid.QLabel("| over-temperature"), i, 2, 1, 2); i+=1
        grid.addWidget(self.status_QC_INH                , i, 0)
        grid.addWidget(QtWid.QLabel("INH")               , i, 1)
        grid.addWidget(QtWid.QLabel("| output inhibited"), i, 2, 1, 2); i+=1
        grid.addWidget(self.status_QC_UNR                , i, 0)
        grid.addWidget(QtWid.QLabel("UNR")               , i, 1)
        grid.addWidget(QtWid.QLabel("| unregulated")     , i, 2, 1, 2); i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)            , i, 0)      ; i+=1
        grid.addWidget(self.status_OC_WTG                , i, 0)
        grid.addWidget(QtWid.QLabel("WTG")               , i, 1)
        grid.addWidget(QtWid.QLabel("| waiting for trigger"), i, 2, 1, 2); i+=1
        grid.addWidget(self.status_OC_CV                 , i, 0)
        grid.addWidget(QtWid.QLabel("CV")                , i, 1)
        grid.addWidget(QtWid.QLabel("| constant voltage"), i, 2, 1, 2); i+=1
        grid.addWidget(self.status_OC_CC                 , i, 0)
        grid.addWidget(QtWid.QLabel("CC")                , i, 1)
        grid.addWidget(QtWid.QLabel("| constant current"), i, 2, 1, 2); i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)            , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Errors")            , i, 0, 1, 2)
        grid.addWidget(self.errors                       , i, 2, 1, 2); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 4)             , i, 0)      ; i+=1
        grid.addWidget(self.pbtn_ackn_errors             , i, 0, 1, 4); i+=1

        hbox = QtWid.QHBoxLayout()
        hbox.addWidget(self.pbtn_save_defaults)
        hbox.addWidget(self.pbtn_reinit)
        grid.addLayout(hbox                              , i, 0, 1, 4); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 4)             , i, 0)      ; i+=1
        grid.addWidget(self.lbl_update_counter           , i, 0, 1, 4); i+=1
        # fmt: on

        grid.setColumnStretch(0, 0)
        grid.setColumnStretch(1, 0)
        grid.setColumnStretch(2, 0)
        grid.setColumnStretch(3, 1)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        # grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
        self.grid = grid

        self.grpb = QtWid.QGroupBox(f"{self.dev.name}")
        self.grpb.setLayout(self.grid)

    # --------------------------------------------------------------------------
    #   update_GUI
    # --------------------------------------------------------------------------

    @Slot()
    def update_GUI(self):
        """NOTE: 'self.dev.mutex' is not being locked, because we are only
        reading 'state' for displaying purposes. We can do this because 'state'
        members are written and read atomicly, with the only exception being
        'all_errors', and it bears no consequences to read wrongly.
        Not locking the mutex might speed up the program.
        """
        if self.dev.is_alive:
            if self.dev.state.ENA_PID:
                self.pbtn_ENA_PID.setChecked(True)
                self.pbtn_ENA_PID.setText("ON")
                self.V_source.setReadOnly(True)
                self.V_source.setText(f"{self.dev.state.V_source:.2f}")
            else:
                self.pbtn_ENA_PID.setChecked(False)
                self.pbtn_ENA_PID.setText("OFF")
                self.V_source.setReadOnly(False)

            if self.dev.state.status_QC_INH:
                self.V_meas.setText("")
                self.I_meas.setText("Inhibited")
                self.I_meas.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
                self.P_meas.setText("")
            else:
                # fmt: off
                self.V_meas.setText(f"{self.dev.state.V_meas:.2f}  V   ")
                self.I_meas.setText(f"{self.dev.state.I_meas:.3f} A   ")
                self.I_meas.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
                self.P_meas.setText(f"{self.dev.state.P_meas:.2f}  W   ")
                # fmt: on

            self.pbtn_ENA_output.setChecked(self.dev.state.ENA_output)
            if self.pbtn_ENA_output.isChecked():
                self.pbtn_ENA_output.setText("Output ON")
            else:
                self.pbtn_ENA_output.setText("Output OFF")

            self.pbtn_ENA_OCP.setChecked(self.dev.state.ENA_OCP)
            if self.pbtn_ENA_OCP.isChecked():
                self.pbtn_ENA_OCP.setText("ON")
            else:
                self.pbtn_ENA_OCP.setText("OFF")

            self.status_QC_OV.setChecked(self.dev.state.status_QC_OV)
            self.status_QC_OC.setChecked(self.dev.state.status_QC_OC)
            self.status_QC_PF.setChecked(self.dev.state.status_QC_PF)
            self.status_QC_OT.setChecked(self.dev.state.status_QC_OT)
            self.status_QC_INH.setChecked(self.dev.state.status_QC_INH)
            self.status_QC_UNR.setChecked(self.dev.state.status_QC_UNR)

            self.status_OC_WTG.setChecked(self.dev.state.status_OC_WTG)
            self.status_OC_CV.setChecked(self.dev.state.status_OC_CV)
            self.status_OC_CC.setChecked(self.dev.state.status_OC_CC)

            self.errors.setReadOnly(self.dev.state.all_errors != [])
            self.errors.setText(";".join(self.dev.state.all_errors))

            self.lbl_update_counter.setText(f"{self.update_counter_DAQ}")
        else:
            self.V_meas.setText("")
            self.I_meas.setText("Offline")
            self.I_meas.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
            self.P_meas.setText("")
            self.grpb.setEnabled(False)

    # --------------------------------------------------------------------------
    #   update_GUI_input_field
    # --------------------------------------------------------------------------

    @Slot()
    @Slot(int)
    def update_GUI_input_field(self, GUI_input_field=GUI_input_fields.ALL):
        if GUI_input_field == GUI_input_fields.OVP_level:
            self.OVP_level.setText(f"{self.dev.state.OVP_level:.2f}")
            self.dev.PID_power.set_output_limits(
                0, self.dev.state.OVP_level * 0.95
            )

        elif GUI_input_field == GUI_input_fields.V_source:
            self.V_source.setText(f"{self.dev.state.V_source:.2f}")

        elif GUI_input_field == GUI_input_fields.I_source:
            self.I_source.setText(f"{self.dev.state.I_source:.3f}")

        elif GUI_input_field == GUI_input_fields.P_source:
            self.P_source.setText(f"{self.dev.state.P_source:.2f}")

        else:
            self.OVP_level.setText(f"{self.dev.state.OVP_level:.2f}")
            self.dev.PID_power.set_output_limits(
                0, self.dev.state.OVP_level * 0.95
            )

            self.V_source.setText(f"{self.dev.state.V_source:.2f}")
            self.I_source.setText(f"{self.dev.state.I_source:.3f}")
            self.P_source.setText(f"{self.dev.state.P_source:.2f}")

    # --------------------------------------------------------------------------
    #   GUI functions
    # --------------------------------------------------------------------------

    @Slot()
    def process_pbtn_ENA_output(self):
        if self.pbtn_ENA_output.isChecked():
            # Clear output protection, if triggered and turn on output
            self.send(self.dev.clear_output_protection_and_turn_on)
        else:
            # Turn off output
            self.send(self.dev.turn_off)

    @Slot()
    def process_pbtn_ENA_PID(self):
        self.dev.state.ENA_PID = self.pbtn_ENA_PID.isChecked()

    @Slot()
    def process_pbtn_ENA_OCP(self):
        self.send(self.dev.set_ENA_OCP, self.pbtn_ENA_OCP.isChecked())

    @Slot()
    def process_pbtn_ackn_errors(self):
        # Lock the dev mutex because string operations are not atomic
        locker = QtCore.QMutexLocker(self.dev.mutex)  # type: ignore
        self.dev.state.all_errors = []
        self.errors.setText("")
        self.errors.setReadOnly(False)  # To change back to regular colors
        locker.unlock()

    @Slot()
    def process_pbtn_reinit(self):
        title = f"Reinitialize {self.dev.name}"
        msg = "Are you sure you want reinitialize the power supply?"
        msgbox = QtWid.QMessageBox()
        msgbox.setIcon(QtWid.QMessageBox.Icon.Question)
        msgbox.setWindowTitle(title)
        msgbox.setText(msg)
        msgbox.setStandardButtons(
            QtWid.QMessageBox.StandardButton.Yes
            | QtWid.QMessageBox.StandardButton.No
        )
        msgbox.setDefaultButton(QtWid.QMessageBox.StandardButton.No)
        reply = msgbox.exec()

        if reply == QtWid.QMessageBox.StandardButton.Yes:
            self.dev.read_config_file()
            self.add_to_jobs_queue(self.dev.reinitialize)
            self.add_to_jobs_queue(
                "signal_GUI_input_field_update", GUI_input_fields.ALL
            )
            self.process_jobs_queue()

            self.dev.state.ENA_PID = False

    @Slot()
    def process_pbtn_save_defaults(self):
        title = f"Save defaults {self.dev.name}"
        msg = (
            "Are you sure you want to save the current values:\n\n"
            "  - Source voltage\n"
            "  - Source current\n"
            "  - Source power\n"
            "  - OVP\n"
            "  - OCP\n\n"
            "as default?\n"
            "These will then automatically be loaded next time."
        )
        msgbox = QtWid.QMessageBox()
        msgbox.setIcon(QtWid.QMessageBox.Icon.Question)
        msgbox.setWindowTitle(title)
        msgbox.setText(msg)
        msgbox.setStandardButtons(
            QtWid.QMessageBox.StandardButton.Yes
            | QtWid.QMessageBox.StandardButton.No
        )
        msgbox.setDefaultButton(QtWid.QMessageBox.StandardButton.No)
        reply = msgbox.exec()

        if reply == QtWid.QMessageBox.StandardButton.Yes:
            if self.dev.write_config_file():
                icon = QtWid.QMessageBox.Icon.Information
                msg = f"Successfully saved to disk:\n{self.dev.path_config}"
            else:
                icon = QtWid.QMessageBox.Icon.Critical
                msg = f"Failed to save to disk:\n{self.dev.path_config}"

            msgbox = QtWid.QMessageBox()
            msgbox.setIcon(icon)
            msgbox.setWindowTitle(title)
            msgbox.setText(msg)
            msgbox.exec()

    @Slot()
    def send_V_source_from_textbox(self):
        try:
            voltage = float(self.V_source.text())
        except (TypeError, ValueError):
            voltage = 0.0
        except Exception as err:
            raise err

        self.add_to_jobs_queue(self.dev.set_V_source, max(voltage, 0))
        self.add_to_jobs_queue(self.dev.query_V_source)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.V_source
        )
        self.process_jobs_queue()

    @Slot()
    def send_I_source_from_textbox(self):
        try:
            current = float(self.I_source.text())
        except (TypeError, ValueError):
            current = 0.0
        except Exception as err:
            raise err

        self.add_to_jobs_queue(self.dev.set_I_source, max(current, 0))
        self.add_to_jobs_queue(self.dev.query_I_source)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.I_source
        )
        self.process_jobs_queue()

    @Slot()
    def set_P_source_from_textbox(self):
        try:
            power = float(self.P_source.text())
        except (TypeError, ValueError):
            power = 0.0
        except Exception as err:
            raise err

        self.dev.state.P_source = max(power, 0)
        self.update_GUI_input_field(GUI_input_fields.P_source)

    @Slot()
    def send_OVP_level_from_textbox(self):
        try:
            OVP_level = float(self.OVP_level.text())
        except (TypeError, ValueError):
            OVP_level = 0.0
        except Exception as err:
            raise err

        self.add_to_jobs_queue(self.dev.set_OVP_level, OVP_level)
        self.add_to_jobs_queue(self.dev.query_OVP_level)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.OVP_level
        )
        self.process_jobs_queue()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Multithreaded PyQt/PySide GUI to interface with an Aim TTi power supply unit
(PSU), QL series II."""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
print(__url__)
# pylint: disable=wrong-import-position, missing-function-docstring

import sys

import qtpy
from qtpy import QtCore, QtGui, QtWidgets as QtWid

import dvg_pyqt_controls as controls
from dvg_devices.Aim_TTi_PSU_protocol_RS232 import Aim_TTi_PSU
from dvg_devices.Aim_TTi_PSU_qdev import Aim_TTi_PSU_qdev

# Show debug info in terminal? Warning: Slow! Do not leave on unintentionally.
DEBUG = False

# ------------------------------------------------------------------------------
#   MainWindow
# ------------------------------------------------------------------------------


class MainWindow(QtWid.QWidget):
    def __init__(self, qdev: Aim_TTi_PSU_qdev, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.setWindowTitle("Aim TTi power supply control")
        self.setGeometry(40, 60, 0, 0)
        self.setFont(QtGui.QFont("Arial", 9))
        self.setStyleSheet(
            controls.SS_TEXTBOX_READ_ONLY
            + controls.SS_GROUP
            + controls.SS_HOVER
        )

        self.pbtn_exit = QtWid.QPushButton("Exit")
        self.pbtn_exit.clicked.connect(self.close)
        self.pbtn_exit.setMinimumHeight(30)

        hbox = QtWid.QHBoxLayout()
        hbox.addWidget(qdev.grpb)
        hbox.addWidget(
            self.pbtn_exit, alignment=QtCore.Qt.AlignmentFlag.AlignTop
        )
        hbox.addStretch(1)

        vbox = QtWid.QVBoxLayout(self)
        vbox.addLayout(hbox)
        vbox.addStretch(1)


# ------------------------------------------------------------------------------
#   Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    # Config file containing COM port address
    PATH_PORT = "config/port_Aim_TTi_PSU.txt"

    # The state of the PSU is polled with this time interval
    DAQ_INTERVAL_MS = 200  # [ms]

    # --------------------------------------------------------------------------
    #   Connect to power supply
    # --------------------------------------------------------------------------

    psu = Aim_TTi_PSU(name="Aim TTi PSU")
    if psu.auto_connect(filepath_last_known_port=PATH_PORT):
        psu.begin()

    # --------------------------------------------------------------------------
    #   Create application
    # --------------------------------------------------------------------------

    main_thread = QtCore.QThread.currentThread()
    if isinstance(main_thread, QtCore.QThread):
        main_thread.setObjectName("MAIN")  # For DEBUG info

    if qtpy.PYQT6 or qtpy.PYSIDE6:
        sys.argv += ["-platform", "windows:darkmode=0"]
    app = QtWid.QApplication(sys.argv)
    app.setStyle("Fusion")

    # --------------------------------------------------------------------------
    #   Set up communication threads for the PSU
    # --------------------------------------------------------------------------

    psu_qdev = Aim_TTi_PSU_qdev(
        dev=psu,
        DAQ_interval_ms=DAQ_INTERVAL_MS,
        debug=DEBUG,
    )
    psu_qdev.start()

    # --------------------------------------------------------------------------
    #   Start the main GUI event loop
    # --------------------------------------------------------------------------

    def about_to_quit():
        print("About to quit")
        app.processEvents()
        psu_qdev.quit()
        psu.close()

    app.aboutToQuit.connect(about_to_quit)
    window = MainWindow(qdev=psu_qdev)
    window.show()

    sys.exit(app.exec())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""RS232 function library for an Aim TTi power supply unit (PSU), QL series II.

Note:
    * Only one channel implemented (channel 1)
    * Limited error reporting
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=broad-except, missing-function-docstring, multiple-statements

import os
import sys
import time
from typing import Union, Tuple
from pathlib import Path

import numpy as np

from dvg_debug_functions import print_fancy_traceback as pft
from dvg_devices.BaseDevice import SerialDevice


class Aim_TTi_PSU(SerialDevice):
    class State:
        """Container for the process and measurement variables."""

        # fmt: off
        V_source: float = 0            # Voltage to be sourced         [V]
        I_source: float = 0            # Current to be sourced (limit) [A]

        V_meas: float = np.nan         # Measured output voltage [V]
        I_meas: float = np.nan         # Measured output current [A]
        P_meas: float = np.nan         # Derived output power    [W]

        OVP_level:  float = np.nan      # Over-voltage protection level [V]
        OCP_level:  float = np.nan      # Over-current protection level [A]
        ENA_output: bool  = False       # Is power output enabled (by software)?

        # Limit Event Status Register (LSR)
        LSR_is_tripped:  bool = False  # True if any LSR tripped, ignoring mode
        LSR_TRIP_AUX:    bool = False  # Aux output trip                 , bit 7
        LSR_MODE_AUX_CC: bool = False  # Aux output constant-current mode, bit 6
        LSR_TRIP_SENSE:  bool = False  # Sense trip                      , bit 5
        LSR_TRIP_OTP:    bool = False  # Over-temperature trip           , bit 4
        LSR_TRIP_OCP:    bool = False  # Over-current trip               , bit 3
        LSR_TRIP_OVP:    bool = False  # Over-voltage trip               , bit 2
        LSR_MODE_CC:     bool = False  # Constant-current mode           , bit 1
        LSR_MODE_CV:     bool = False  # Constant-voltage mode           , bit 0
        # fmt: on

    class Config:
        # fmt: off
        V_source:  float = 10   # Voltage to be sourced         [V]
        I_source:  float = 0.5  # Current to be sourced (limit) [A]
        OVP_level: float = 12   # Over-voltage protection level [V]
        OCP_level: float = 1    # Over-current protection level [A]
        # fmt: on

    # --------------------------------------------------------------------------
    #   __init__
    # --------------------------------------------------------------------------

    def __init__(
        self,
        name: str = "PSU",
        long_name: str = "Aim TTi power supply",
        path_config: str = (os.getcwd() + "/config/settings_Aim_TTi_PSU.txt"),
        connect_to_serial_number=None,
    ):
        super().__init__(name=name, long_name=long_name)

        # Default serial settings
        self.serial_settings = {
            "baudrate": 9600,  # baudrate gets ignored
            "timeout": 0.2,
            "write_timeout": 0.2,
        }
        self.set_read_termination("\r\n")
        self.set_write_termination("\r\n")

        self.set_ID_validation_query(
            ID_validation_query=self.ID_validation_query,
            valid_ID_broad=True,
            valid_ID_specific=connect_to_serial_number,
        )

        # Container for the process and measurement variables
        self.state = self.State()
        self.config = self.Config()

        self.idn_str = ""  # Identity response of the device
        self.serial_str = ""  # Serial number of the device
        self.model_str = ""  # Model of the device

        # Location of the configuration file
        self.path_config = Path(path_config)

    # --------------------------------------------------------------------------
    #   write_and_wait_for_opc
    # --------------------------------------------------------------------------

    def write_and_wait_for_opc(
        self,
        msg: Union[str, bytes],
        raises_on_timeout: bool = False,
    ) -> bool:
        """For proper synchronization we have to wait for the Operation Complete
        status of the device for the majority of the `set` commands. Hence this
        method.
        """
        success = self.write(msg=msg, raises_on_timeout=raises_on_timeout)
        if success:
            self.wait_for_OPC()  # Crucial!

        return success

    # --------------------------------------------------------------------------
    #   ID_validation_query
    # --------------------------------------------------------------------------

    def ID_validation_query(self) -> Tuple[bool, str]:
        success = self.query_IDN()
        return success, self.serial_str

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(self) -> bool:
        """This function should run directly after having established a
        connection to the device.

        Returns: True if successful, False otherwise.
        """
        if not self.is_alive:
            print("ERROR: Device is not connected yet or already closed.")
            return False

        success = self.set_LSE(value=255)  # Report on all trips and modes
        success &= self.query_OVP_level()
        success &= self.query_OCP_level()
        success &= self.query_V_source()
        success &= self.query_I_source()
        success &= self.query_LSR()

        return success

    # --------------------------------------------------------------------------
    #   flush_serial_input_and_output
    # --------------------------------------------------------------------------

    def flush_serial_input_and_output(self):
        # Flush the serial input and output buffers
        self.ser.flushInput()  # type: ignore
        self.ser.flushOutput()  # type: ignore

    # --------------------------------------------------------------------------
    #   reinitialize
    # --------------------------------------------------------------------------

    def reinitialize(self) -> bool:
        """Reinitialize the PSU, including clear and reset

        Returns: True if all messages were sent and received successfully,
            False otherwise.
        """

        if not self.is_alive:
            print("ERROR: Device is not connected yet or already closed.")
            return False

        self.flush_serial_input_and_output()

        success = self.clear_and_reset()
        success &= self.set_OVP_level(self.config.OVP_level)
        success &= self.set_OCP_level(self.config.OCP_level)
        success &= self.set_V_source(self.config.V_source)
        success &= self.set_I_source(self.config.I_source)

        success &= self.set_LSE(value=255)  # Report on all trips and modes
        success &= self.query_OVP_level()
        success &= self.query_OCP_level()
        success &= self.query_V_source()
        success &= self.query_I_source()
        success &= self.query_LSR()

        return success

    # --------------------------------------------------------------------------
    #   query_IDN
    # --------------------------------------------------------------------------

    def query_IDN(self) -> bool:
        """Query the identity, serial and model number and store it in the class
        members.

        Returns: True if successful, False otherwise.
        """
        success, reply = self.query("*idn?")
        if (
            success
            and isinstance(reply, str)
            and reply[:19] == "THURLBY THANDAR, QL"
        ):
            self.idn_str = reply
            self.serial_str = reply.split(",")[2].strip()
            self.model_str = reply.split(",")[1].strip()
            return True

        self.idn_str = ""
        self.serial_str = ""
        self.model_str = ""
        return False

    # --------------------------------------------------------------------------
    #   System status related
    # --------------------------------------------------------------------------

    def clear_and_reset(self) -> bool:
        """Clear device status and reset. Return when this operation has
        completed on the device. Blocking.

        Returns: True if the message was sent successfully, False otherwise.
        """

        if not self.is_alive:
            print("ERROR: Device is not connected yet or already closed.")
            return False

        # Send clear and reset
        success = self.write("*cls;*rst")
        if success:
            # For some reason '*opc?' will not wait for operation complete
            # after a '*cls;*rst'. Hence, we will sleep instead.
            time.sleep(0.2)  # Crucial!

        return success

    def wait_for_OPC(self) -> bool:
        """'Operation complete' query, used for event synchronization. Will wait
        for all device operations to complete or until a timeout is triggered.
        Blocking.

        Returns: True if successful, False otherwise.
        """
        # Returns an ASCII "1" when all pending overlapped operations have been
        # completed.
        success, reply = self.query("*opc?")
        if success and reply == "1":
            return True

        pft(f"Warning: *opc? timed out at device {self.name}")
        return False

    def set_LSE(self, value: int, channel: int = 1) -> bool:
        """Set the value of the Limit Event Status Enable Register (LSE).

        Args:
            value (int): (0-255)
                * bit 7: Aux output trip
                * bit 6: Aux output constant-current mode
                * bit 5: Sense trip
                * bit 4: Over-temperature trip
                * bit 3: Over-current trip
                * bit 2: Over-voltage trip
                * bit 1: Constant-current mode
                * bit 0: Constant-voltage mode

        Returns: True if successful, False otherwise.
        """
        return self.write_and_wait_for_opc(f"LSE{channel} {value}")

    def query_LSR(self, verbose: bool = False, channel: int = 1) -> bool:
        """Query and parse the Limit Event Status Register (LSR). This holds
        the protection trips and output modes.

        Returns: True if successful, False otherwise.
        """
        # We will ignore the first read-out, because it apparently always lags
        # one behind.
        self.query(f"LSR{channel}?")

        success, reply = self.query(f"LSR{channel}?")
        if success and isinstance(reply, str):
            # fmt: off
            status_code = int(reply)
            self.state.LSR_TRIP_AUX    = bool(status_code & 128)
            self.state.LSR_MODE_AUX_CC = bool(status_code & 64)
            self.state.LSR_TRIP_SENSE  = bool(status_code & 32)
            self.state.LSR_TRIP_OTP    = bool(status_code & 16)
            self.state.LSR_TRIP_OCP    = bool(status_code & 8)
            self.state.LSR_TRIP_OVP    = bool(status_code & 4)
            self.state.LSR_MODE_CC     = bool(status_code & 2)
            self.state.LSR_MODE_CV     = bool(status_code & 1)

            self.state.LSR_is_tripped = (
                self.state.LSR_TRIP_AUX
                or self.state.LSR_TRIP_SENSE
                or self.state.LSR_TRIP_OTP
                or self.state.LSR_TRIP_OCP
                or self.state.LSR_TRIP_OVP
            )

            if verbose:
                if self.state.LSR_TRIP_AUX    : print("  AUX TRIP")
                if self.state.LSR_MODE_AUX_CC : print("  AUX CC mode")
                if self.state.LSR_TRIP_SENSE  : print("  SENSE TRIP")
                if self.state.LSR_TRIP_OTP    : print("  OVER-TEMPERATURE TRIP")
                if self.state.LSR_TRIP_OCP    : print("  OVER-CURRENT TRIP")
                if self.state.LSR_TRIP_OVP    : print("  OVER-VOLTAGE TRIP")
                if self.state.LSR_MODE_CC     : print("  CC mode")
                if self.state.LSR_MODE_CV     : print("  CV mode")
            # fmt: on

        return success

    # --------------------------------------------------------------------------
    #   Protection related
    # --------------------------------------------------------------------------

    def reset_trips(self) -> bool:
        """Attempt to clear all trip conditions from all outputs.

        Returns: True if successful, False otherwise.
        """
        return self.write_and_wait_for_opc("triprst")

    def set_OVP_level(self, voltage_V, channel: int = 1) -> bool:
        """Returns: True if the message was sent successfully, False otherwise."""
        try:
            voltage_V = float(voltage_V)
        except (ValueError, TypeError):
            voltage_V = 0.0
        except Exception as e:
            raise e

        if self.write_and_wait_for_opc(f"OVP{channel} {voltage_V}"):
            self.state.OVP_level = voltage_V
            return True

        return False

    def set_OCP_level(self, current_A, channel: int = 1) -> bool:
        """Returns: True if the message was sent successfully, False otherwise."""
        try:
            current_A = float(current_A)
        except (ValueError, TypeError):
            current_A = 0.0
        except Exception as e:
            raise e

        if self.write_and_wait_for_opc(f"OCP{channel} {current_A}"):
            self.state.OCP_level = current_A
            return True

        return False

    def query_OVP_level(self, channel: int = 1) -> bool:
        """Returns: True if successful, False otherwise."""
        success, reply = self.query(f"OVP{channel}?")
        if success:
            if isinstance(reply, str) and reply[:3] == f"VP{channel}":
                self.state.OVP_level = float(reply[4:])
                return True
            else:
                pft(f"Received incorrect reply: {reply}")

        return False

    def query_OCP_level(self, channel: int = 1) -> bool:
        """Returns: True if successful, False otherwise."""
        success, reply = self.query(f"OCP{channel}?")
        if success:
            if isinstance(reply, str) and reply[:3] == f"CP{channel}":
                self.state.OCP_level = float(reply[4:])
                return True
            else:
                pft(f"Received incorrect reply: {reply}")

        return False

    # --------------------------------------------------------------------------
    #   Output related
    # --------------------------------------------------------------------------

    def reset_trips_and_turn_on(self, channel: int = 1) -> bool:
        """Returns: True if the message was sent successfully, False otherwise."""
        if self.write_and_wait_for_opc(f"triprst;*opc;OP{channel} 1"):
            self.state.ENA_output = True
            return True

        return False

    def turn_on(self, channel: int = 1) -> bool:
        """Returns: True if the message was sent successfully, False otherwise."""
        return self.set_ENA_output(ENA=True, channel=channel)

    def turn_off(self, channel: int = 1) -> bool:
        """Returns: True if the message was sent successfully, False otherwise."""
        return self.set_ENA_output(ENA=False, channel=channel)

    def set_ENA_output(self, ENA: bool, channel: int = 1) -> bool:
        """Returns: True if the message was sent successfully, False otherwise."""
        if self.write_and_wait_for_opc(f"OP{channel} {ENA}"):
            self.state.ENA_output = ENA
            return True

        return False

    def set_V_source(self, voltage_V, channel: int = 1) -> bool:
        """Returns: True if the message was sent successfully, False otherwise."""
        try:
            voltage_V = float(voltage_V)
        except (ValueError, TypeError):
            voltage_V = 0.0
        except Exception as e:
            raise e

        if self.write_and_wait_for_opc(f"V{channel} {voltage_V:.3f}"):
            self.state.V_source = voltage_V
            return True

        return False

    def set_I_source(self, current_A, channel: int = 1) -> bool:
        """Returns: True if the message was sent successfully, False otherwise."""
        try:
            current_A = float(current_A)
        except (ValueError, TypeError):
            current_A = 0.0
        except Exception as e:
            raise e

        if self.write_and_wait_for_opc(f"I{channel} {current_A:.4f}"):
            self.state.I_source = current_A
            return True

        return False

    def query_ENA_output(self, channel: int = 1) -> bool:
        """Returns: True if the query was received successfully, False otherwise."""
        success, reply = self.query(f"OP{channel}?")
        if success:
            if isinstance(reply, str) and reply.isnumeric():
                self.state.ENA_output = bool(int(reply))
                return True

            pft(f"Received incorrect reply: {reply}")
            self.flush_serial_input_and_output()

        return False

    def query_V_source(self, channel: int = 1) -> bool:
        """Returns: True if the query was received successfully, False otherwise."""
        success, reply = self.query(f"V{channel}?")
        if success:
            if isinstance(reply, str) and reply[:2] == f"V{channel}":
                self.state.V_source = float(reply[3:])
                return True

            pft(f"Received incorrect reply: {reply}")
            self.flush_serial_input_and_output()

        return False

    def query_V_meas(self, channel: int = 1) -> bool:
        """Returns: True if the query was received successfully, False otherwise."""
        success, reply = self.query(f"V{channel}O?")
        if success:
            if isinstance(reply, str) and reply[-1:] == "V":
                self.state.V_meas = float(reply[:-1])
                self.state.P_meas = self.state.I_meas * self.state.V_meas
                return True

            pft(f"Received incorrect reply: {reply}")
            self.flush_serial_input_and_output()

        return False

    def query_I_source(self, channel: int = 1) -> bool:
        """Returns: True if the query was received successfully, False otherwise."""
        success, reply = self.query(f"I{channel}?")
        if success:
            if isinstance(reply, str) and reply[:2] == f"I{channel}":
                self.state.I_source = float(reply[3:])
                return True

            pft(f"Received incorrect reply: {reply}")
            self.flush_serial_input_and_output()

        return False

    def query_I_meas(self, channel: int = 1) -> bool:
        """Returns: True if the query was received successfully, False otherwise."""
        success, reply = self.query(f"I{channel}O?")
        if success:
            if isinstance(reply, str) and reply[-1:] == "A":
                self.state.I_meas = float(reply[:-1])
                self.state.P_meas = self.state.I_meas * self.state.V_meas
                return True

            pft(f"Received incorrect reply: {reply}")
            self.flush_serial_input_and_output()

        return False

    # --------------------------------------------------------------------------
    #   Speed tests for debugging
    # --------------------------------------------------------------------------

    def speed_test(self):
        """Results: Each iteration takes 80 ms to finish."""
        self.turn_off()  # Disable output for safety

        tic = time.perf_counter()
        for i in range(100):
            print(f"i {time.perf_counter() - tic:.3f}")
            self.set_V_source(i % 10)

        print(time.perf_counter() - tic)
        self.report()

    def speed_test2(self):
        """Results: Each iteration takes 8 ms to finish."""

        tic = time.perf_counter()
        for i in range(100):
            print(
                f"{i} {time.perf_counter() - tic:.3f} "
                f"{self.state.V_meas:.3f} {self.state.I_meas:.3f}"
            )
            self.query_V_meas()
            self.query_I_meas()

        print(time.perf_counter() - tic)
        self.report()

    # --------------------------------------------------------------------------
    #   report
    # --------------------------------------------------------------------------

    def report(self, channel: int = 1):
        """Report to the terminal."""
        self.query_LSR(verbose=True, channel=channel)
        self.query_ENA_output(channel=channel)
        self.query_OVP_level(channel=channel)
        self.query_OCP_level(channel=channel)
        self.query_V_source(channel=channel)
        self.query_I_source(channel=channel)
        self.query_V_meas(channel=channel)
        self.query_I_meas(channel=channel)

        print("")
        print(
            f"  {('ON' if self.state.ENA_output else 'OFF'):-3s}       "
            f"  {self.state.OVP_level:4.1f} OVP   "
            f"  {self.state.OCP_level:4.2f} OCP"
        )
        print("  " + chr(0x2014) * 46)

        print(
            f"  Source "
            f"     {self.state.V_source:6.3f} V"
            f"     {self.state.I_source:6.4f} A"
        )
        print("  " + chr(0x2014) * 46)
        print(
            f"  Measure "
            f"    {self.state.V_meas:6.3f} V "
            f"    {self.state.I_meas:6.4f} A "
            f"    {self.state.P_meas:6.3f} W"
        )

    # --------------------------------------------------------------------------
    #   read_config_file
    # --------------------------------------------------------------------------

    def read_config_file(self) -> bool:
        """Try to open the config textfile containing:
            * V_source   # Voltage to be sourced         [V]
            * I_source   # Current to be sourced (limit) [A]
            * OVP_level  # Over-voltage protection level [V]
            * OCP_level  # Over-current protection level [A]
        Do not panic if the file does not exist or cannot be read.

        Returns: True if successful, False otherwise.
        """
        if self.path_config.is_file():
            try:
                with self.path_config.open(encoding="utf8") as f:
                    self.config.V_source = float(f.readline().strip())
                    self.config.I_source = float(f.readline().strip())
                    self.config.OVP_level = float(f.readline().strip())
                    self.config.OCP_level = float(f.readline().strip())

                return True
            except Exception:
                pass  # Do not panic and remain silent

        return False

    # --------------------------------------------------------------------------
    #   write_config_file
    # --------------------------------------------------------------------------

    def write_config_file(self) -> bool:
        """Try to write the config textfile containing:
             * V_source   # Voltage to be sourced         [V]
             * I_source   # Current to be sourced (limit) [A]
             * OVP_level  # Over-voltage protection level [V]
             * OCP_level  # Over-current protection level [A]
        Do not panic if the file does not exist or cannot be read.

        Returns: True if successful, False otherwise.
        """

        if not self.path_config.parent.is_dir():
            # Subfolder does not exists yet. Create.
            try:
                self.path_config.parent.mkdir()
            except Exception:
                pass  # Do not panic and remain silent

        try:
            # Write the config file
            self.path_config.write_text(
                f"{self.state.V_source:.3f}\n"
                f"{self.state.I_source:.3f}\n"
                f"{self.state.OVP_level:.1f}\n"
                f"{self.state.OCP_level:.2f}",
                encoding="utf8",
            )
        except Exception:
            pass  # Do not panic and remain silent
        else:
            return True

        return False


# ------------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    # Serial number of the Aim TTi PSU to connect to.
    # Set to '' or None to connect to any Aim TTi PSU.
    # SERIAL_PSU = "527254"
    SERIAL_PSU = None

    # Path to the textfile containing the (last used) RS232 port
    PATH_PORT = "config/port_Aim_TTi_PSU.txt"

    # Create connection to Aim TTi PSU over RS232
    psu = Aim_TTi_PSU(connect_to_serial_number=SERIAL_PSU)

    if psu.auto_connect(PATH_PORT):
        psu.begin()  # Retrieve necessary parameters
        print(f"  IDN: {psu.idn_str}")

        psu.report()
        psu.close()

    sys.exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Provides base class SerialDevice(), offering higher-level general I/O methods
for a serial device, like autoconnect. Instances of this class will tie in
nicely with :class:`dvg_qdeviceio.QDeviceIO`.

These base classes are meant to be inherited into your own specific *Device*
class.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=broad-except

import sys
import time
from typing import Union, Tuple, Callable
from pathlib import Path

# Use of `ast.literal_eval` got removed in v0.2.2 because it chokes on `nan`
# from ast import literal_eval

import serial
import serial.tools.list_ports

from dvg_debug_functions import dprint, ANSI, print_fancy_traceback as pft


class SerialDevice:
    """This class provides higher-level general I/O methods for a serial device,
    by wrapping the excellent `pySerial
    <https://pyserial.readthedocs.io/en/latest/pyserial.html>`_ library.

    The following functionality is offered:

    * TODO: mention `write()`, `query()`, `query_ascii_values()`,
      `query_bytes()`, `readline()` and `close()`.

    * Scanning over all serial ports to autoconnect to the desired serial
      device, based on the device's reply to a validation query.

    * Autoconnecting to the device, based on the last known successful port that
      gets written to a configuration textfile.

    Instances of this class will tie in nicely with
    :class:`dvg_qdeviceio.QDeviceIO`.

    Args:
        long_name (:obj:`str`, optional):
            Long display name of the device in a general sense. E.g.,
            `"Keysight N8700 PSU"` or `"Arduino M0 Pro"`.

            Default: `"Serial Device"`

        name (:obj:`str`, optional):
            Short display name for the device. E.g., `"PSU_1"` or `"blinker"`.

            Default: `"Dev_1"`

    .. rubric:: Attributes:

    Attributes:
        long_name (:obj:`str`):
            Long display name of the device in a general sense. E.g,
            `"Keysight N8700 PSU"` or `"Arduino M0 Pro"`.

        name (:obj:`str`):
            Short display name for the device. E.g., `"PSU_1"` or `"blinker"`.

        serial_settings (:obj:`dict`):
            Dictionary of keyword arguments to be passed directly to
            :class:`serial.Serial` at initialization of the serial port when
            trying to connect. Do not specify `port` in this dictionary as it
            will be supplied by this class's machinery.

            Default: `{"baudrate": 9600, "timeout": 2, "write_timeout": 2}`

        ser (:class:`serial.Serial` | :obj:`None`):
            Will be set to a :class:`serial.Serial` device instance when a
            connection has been established. Otherwise: :obj:`None`.

        is_alive (:obj:`bool`):
            Is the connection alive? I.e., Can we communicate?
    """

    def __init__(
        self,
        name="Dev_1",
        long_name="Serial Device",
    ):
        self.long_name = long_name
        self.name = name

        # Default serial settings
        self.serial_settings = {
            # Defaults from `serial.Serial`
            "bytesize": 8,
            "parity": "N",
            "stopbits": 1,
            "xonxoff": False,
            "rtscts": False,
            "dsrdtr": False,
            "inter_byte_timeout": None,
            "exclusive": None,
            # Edited
            "baudrate": 9600,
            "timeout": 2,
            "write_timeout": 2,
        }

        # Termination characters, must always be of type `bytes`.
        self._read_termination: bytes = "\n".encode()
        self._write_termination: bytes = "\n".encode()

        # Wait time during :meth:`query` in case there is no read termination
        # character set. We have to wait long enough to make sure the device has
        # had ample time to send out the reply to the serial-input buffer of
        # the host PC, which we will then read completely in one go.
        self._query_wait_time = 0.1  # [s]

        # See :meth:`set_ID_validation_query`
        self._ID_validation_query = None
        self._valid_ID_broad = None
        self._valid_ID_specific = None
        # Whenever `connect_at_port()` is performing the user-supplied
        # ID_validation_query, any query inside of it /must/ raise an exception
        # on timeout. Instead of relying on the user to set the argument
        # `raises_on_timeout=True` when invoking `query()`, we will use below
        # flag instead to enforce the raise.
        self._force_query_to_raise_on_timeout = False

        self.ser: serial.Serial = None  # type: ignore
        self.is_alive = False

    # --------------------------------------------------------------------------
    #   set_read_termination
    # --------------------------------------------------------------------------

    def set_read_termination(
        self,
        termination: Union[str, bytes, None],
        query_wait_time: float = 0.1,
    ):
        """Set the termination character(s) for serial read.

        Args:
            termination (:obj:`str` | :obj:`bytes` | :obj:`None`):
                Termination character(s). When set to :obj:`None` or empty the
                I/O operation :meth:`query` will wait ``query_wait_time``
                seconds before reading out the complete serial-input buffer,
                which by then should contain the device's reply to the query.

            query_wait_time (:obj:`float`, optional):
                See above.

                Default: :const:`0.1`
        """
        if termination is None:
            termination = b""

        if isinstance(termination, str):
            termination = termination.encode()

        self._read_termination = bytes(termination)
        self._query_wait_time = query_wait_time

    # --------------------------------------------------------------------------
    #   set_write_termination
    # --------------------------------------------------------------------------

    def set_write_termination(self, termination: Union[str, bytes, None]):
        """Set the termination character(s) for serial write.

        Args:
            termination (:obj:`str` | :obj:`bytes` | :obj:`None`):
                Termination character(s).
        """
        if termination is None:
            termination = b""

        if isinstance(termination, str):
            termination = termination.encode()

        self._write_termination = bytes(termination)

    # --------------------------------------------------------------------------
    #   set_ID_validation_query
    # --------------------------------------------------------------------------

    def set_ID_validation_query(
        self,
        ID_validation_query: Callable[[], tuple],
        valid_ID_broad: object,
        valid_ID_specific: object = None,
    ):
        """When this method **is not** called, then the following default scheme
        applies:

            During :meth:`connect_at_port`, :meth:`scan_ports` or
            :meth:`auto_connect` a succesful connection to *any* serial device
            will be accepted and be stored in class member :attr:`ser`.

        When this method **is** called, then the following scheme applies:

            During :meth:`connect_at_port`, :meth:`scan_ports` or
            :meth:`auto_connect` a connection to a *desired* device will be
            attempted by performing a query for a device ID over the serial
            connection. The serial connection will be accepted and be stored in
            class member :attr:`ser` whenever the following scheme returns
            succesful:

        Args:
            ID_validation_query (:obj:`~collections.abc.Callable` [[], :obj:`tuple`]):
                Reference to a function to perform an ID validation query
                on the device when connecting. The function should take zero
                arguments and return a tuple consisting of two objects, as will
                be explained further down. Only when the outcome of the
                validation is successful, will the connection be accepted and
                remain open. Otherwise, the connection will be automatically
                closed again.

                The device's reply on the validation query will be tested
                against the two other arguments: ``valid_ID_broad`` (required)
                and ``valid_ID_specific`` (optional).

                The *broad* reply can be used to allow connections to a device
                of a certain manufacturer and model. E.g., in a response to an
                ``*idn?`` validation query, one can test a part of the full
                reply -- say, `"THURLBY THANDAR, QL355TP, 279730, 1.00 – 1.00"`
                -- against `"THURLBY THANDAR, QL"` to allow connections to *any*
                Thurlby Thandar QL-series power supply.

                The *specific* reply can be used to narrow down to a specific
                device, once it has passed the *broad* validation test. In the
                example above, one could test against the series number
                `"279730"`, for instance. When argument ``valid_ID_specific``
                is not supplied, any *broad* match will be accepted as
                connection.

                The function to be passed, being defined as a class method
                inside of a derived :class:`SerialDevice` class, should have a
                general form like:

                .. code-block:: python

                    def my_ID_validation_query(self) -> tuple[str, str]:
                        # Expected: reply = "THURLBY THANDAR, QL355TP, 279730, 1.00 – 1.00"
                        _success, reply = self.query("*idn?")
                        reply_broad = ans[:19]                  # "THURLBY THANDAR, QL"
                        reply_specific = ans.split(",")[2]      # "279730", i.e. serial number

                        return reply_broad, reply_specific

                When ``ID_validation_query`` is set to :obj:`None`, no
                validation will take place and *any* successful connection will
                be accepted and remain open.

            valid_ID_broad (:obj:`object`):
                Reply to be broadly matched when a function reference is being
                passed onto ``ID_validation_query``.

            valid_ID_specific (:obj:`object`, optional):
                Reply to be specifically matched when a function reference is
                being passed onto ``ID_validation_query``. Note: When set
                to :obj:`None`, any connection that is *broadly* matched will be
                accepted and remain open.

                Default: :obj:`None`
        """
        self._ID_validation_query = ID_validation_query
        self._valid_ID_broad = valid_ID_broad
        self._valid_ID_specific = valid_ID_specific

    # --------------------------------------------------------------------------
    #   readline
    # --------------------------------------------------------------------------

    def readline(
        self,
        raises_on_timeout: bool = False,
        returns_ascii: bool = True,
    ) -> Tuple[bool, Union[str, bytes, None]]:
        """Listen to the Arduino for incoming data. This method is blocking
        and returns when a full line has been received or when the serial read
        timeout has expired.

        Args:
            raises_on_timeout (:obj:`bool`, optional):
                Should an exception be raised when a read timeout occurs?

                Default: :const:`False`

            returns_ascii (:obj:`bool`, optional):
                When set to :const:`True` the device's reply will be returned as
                an ASCII string. Otherwise, it will return as bytes.

                Default: :const:`True`

        Returns:
            :obj:`tuple`:
                success (:obj:`bool`):
                    True if successful, False otherwise.

                reply (:obj:`str` | :obj:`bytes` | :obj:`None`):
                    Reply received from the device, either as ASCII string
                    (default) or as bytes when ``returns_ascii`` was set to
                    :const:`False`. :obj:`None` if unsuccessful.
        """

        try:
            reply = self.ser.readline()
        except serial.SerialException as err:
            # NOTE: The Serial library does not throw an exception when it
            # times out in `read`, only when it times out in `write`! We
            # will check for zero received bytes as indication for a read
            # timeout, later. See:
            # https://stackoverflow.com/questions/10978224/serialtimeoutexception-in-python-not-working-as-expected
            pft(err)
            return False, None
        except Exception as err:
            pft(err)
            return False, None

        if len(reply) == 0:
            if raises_on_timeout:
                raise serial.SerialException(
                    "Received 0 bytes. Read probably timed out."
                )
            else:
                pft("Received 0 bytes. Read probably timed out.")
                return False, None

        if returns_ascii:
            try:
                reply = reply.decode("utf8").strip()
            except Exception as err:
                pft(err)
                return False, None

        return True, reply

    # --------------------------------------------------------------------------
    #   write
    # --------------------------------------------------------------------------

    def write(
        self, msg: Union[str, bytes], raises_on_timeout: bool = False
    ) -> bool:
        """Send a message to the serial device.

        Args:
            msg (:obj:`str` | :obj:`bytes`):
                ASCII string or bytes to be sent to the serial device.

            raises_on_timeout (:obj:`bool`, optional):
                Should an exception be raised when a write timeout occurs?

                Default: :const:`False`

        Returns:
            True if successful, False otherwise.
        """
        if not self.is_alive:
            pft("Device is not connected yet or already closed.", 3)
            return False  # --> leaving

        if isinstance(msg, str):
            msg = msg.encode()

        try:
            self.ser.write(bytes(msg) + self._write_termination)
        except (
            serial.SerialTimeoutException,
            serial.SerialException,
        ) as err:
            if raises_on_timeout:
                raise err  # --> leaving

            pft(err, 3)
            return False  # --> leaving
        except Exception as err:
            pft(err, 3)
            sys.exit(0)  # --> leaving

        return True

    # --------------------------------------------------------------------------
    #   query
    # --------------------------------------------------------------------------

    def query(
        self,
        msg: Union[str, bytes],
        raises_on_timeout: bool = False,
        returns_ascii: bool = True,
    ) -> Tuple[bool, Union[str, bytes, None]]:
        """Send a message to the serial device and subsequently read the reply.

        Args:
            msg (:obj:`str` | :obj:`bytes`):
                ASCII string or bytes to be sent to the serial device.

            raises_on_timeout (:obj:`bool`, optional):
                Should an exception be raised when a write or read timeout
                occurs?

                Default: :const:`False`

            returns_ascii (:obj:`bool`, optional):
                When set to :const:`True` the device's reply will be returned as
                an ASCII string. Otherwise, it will return as bytes.

                TODO & NOTE: ASCII is a misnomer. The returned reply will be
                UTF-8 encoded, not ASCII. Need to fix the argument name somehow,
                without breaking code elsewhere.

                Default: :const:`True`

        Returns:
            :obj:`tuple`:
                success (:obj:`bool`):
                    True if successful, False otherwise.

                reply (:obj:`str` | :obj:`bytes` | :obj:`None`):
                    Reply received from the device, either as ASCII string
                    (default) or as bytes when ``returns_ascii`` was set to
                    :const:`False`. :obj:`None` if unsuccessful.
        """

        # Always ensure that a timeout exception is raised when coming from
        # :meth:`connect_at_port`.
        if self._force_query_to_raise_on_timeout:
            raises_on_timeout = True

        # Send query
        if not self.write(msg, raises_on_timeout=raises_on_timeout):
            return (False, None)  # --> leaving

        # Read reply
        try:
            if self._read_termination == b"":
                self.ser.flush()
                time.sleep(self._query_wait_time)
                reply = self.ser.read(self.ser.in_waiting)
            else:
                reply = self.ser.read_until(self._read_termination)
        except serial.SerialException as err:
            # Note: The Serial library does not throw an exception when it
            # times out in `read`, only when it times out in `write`! We
            # will check for zero received bytes as indication for a read
            # timeout, later. See:
            # https://stackoverflow.com/questions/10978224/serialtimeoutexception-in-python-not-working-as-expected
            pft(err, 3)
            return (False, None)  # --> leaving
        except Exception as err:
            pft(err, 3)
            sys.exit(0)  # --> leaving

        if len(reply) == 0:
            if raises_on_timeout:
                raise serial.SerialException(
                    "Received 0 bytes. Read probably timed out."
                )  # --> leaving

            pft("Received 0 bytes. Read probably timed out.", 3)
            return (False, None)  # --> leaving

        if returns_ascii:
            try:
                reply = reply.decode("utf8").strip()
            except UnicodeDecodeError as err:
                pft(err, 3)
                return (False, None)  # --> leaving
            except Exception as err:
                pft(err, 3)
                sys.exit(0)  # --> leaving

        return (True, reply)

    # --------------------------------------------------------------------------
    #   query_bytes
    # --------------------------------------------------------------------------

    def query_bytes(
        self,
        msg: bytes,
        N_bytes_to_read: int,
        raises_on_timeout: bool = False,
    ) -> Tuple[bool, Union[bytes, None]]:
        """Send a message as bytes to the serial device and subsequently read
        the reply. Will block until reaching ``N_bytes_to_read`` or a read
        timeout occurs.

        Args:
            msg (:obj:`bytes`):
                Bytes to be sent to the serial device.

            N_bytes_to_read (:obj:`int`):
                Number of bytes to read.

            raises_on_timeout (:obj:`bool`, optional):
                Should an exception be raised when a write or read timeout
                occurs?

                Default: :const:`False`

        Returns:
            :obj:`tuple`:
                success (:obj:`bool`):
                    True when ``N_bytes_to_read`` bytes are indeed read within
                    the timeout, False otherwise.

                reply (:obj:`bytes` | :obj:`None`):
                    Reply received from the device as bytes.

                    If ``success`` is False and 0 bytes got returned from the
                    device, then ``reply`` will be :obj:`None`.
                    If ``success`` is False because the read timed out and too
                    few bytes got returned, ``reply`` will contain the bytes
                    read so far.
        """

        # Always ensure that a timeout exception is raised when coming from
        # :meth:`connect_at_port`.
        if self._force_query_to_raise_on_timeout:
            raises_on_timeout = True

        # Send query
        if not self.write(msg, raises_on_timeout=raises_on_timeout):
            return (False, None)  # --> leaving

        # Read reply
        try:
            if N_bytes_to_read > 0:
                reply = self.ser.read(N_bytes_to_read)
            else:
                reply = b""
                self.ser.flush()
        except serial.SerialException as err:
            # Note: The Serial library does not throw an exception when it
            # times out in `read`, only when it times out in `write`! We
            # will check for zero received bytes as indication for a read
            # timeout, later. See:
            # https://stackoverflow.com/questions/10978224/serialtimeoutexception-in-python-not-working-as-expected
            pft(err, 3)
            return (False, None)  # --> leaving
        except Exception as err:
            pft(err, 3)
            sys.exit(0)  # --> leaving

        if (N_bytes_to_read > 0) and (len(reply) == 0):
            if raises_on_timeout:
                raise serial.SerialException(
                    "Received 0 bytes. Read probably timed out."
                )  # --> leaving

            pft("Received 0 bytes. Read probably timed out.", 3)
            return (False, None)  # --> leaving

        if N_bytes_to_read != len(reply):
            if raises_on_timeout:
                raise serial.SerialException(
                    "Received too few bytes. Read probably timed out."
                )  # --> leaving

            pft("Received too few bytes. Read probably timed out.", 3)
            return (False, reply)  # --> leaving

        return (True, reply)

    # --------------------------------------------------------------------------
    #   query_ascii_values
    # --------------------------------------------------------------------------

    def query_ascii_values(
        self,
        msg: str,
        delimiter="\t",
        raises_on_timeout: bool = False,
    ) -> Tuple[bool, list]:
        r"""Send a message to the serial device and subsequently read the reply.
        Expects a reply in the form of an ASCII string containing a list of
        numeric values, separated by a delimiter. These values will be parsed
        into a list and returned.

        Args:
            msg (:obj:`str`):
                ASCII string to be sent to the serial device.

            delimiter (:obj:`str`, optional):
                Delimiter used in the device's reply.

                Default: `"\\t"`

            raises_on_timeout (:obj:`bool`, optional):
                Should an exception be raised when a write or read timeout
                occurs?

                Default: :const:`False`

        Returns:
            :obj:`tuple`:
                success (:obj:`bool`):
                    True if successful, False otherwise.

                reply_list (:obj:`list`):
                    Reply received from the device and parsed into a list of
                    separate values. The list is empty if unsuccessful.
        """
        success, reply = self.query(
            msg, raises_on_timeout=raises_on_timeout, returns_ascii=True
        )

        if not success or not isinstance(reply, str):
            return (False, [])  # --> leaving

        try:
            # NOTE: `ast.literal_eval` chokes when it receives 'nan' so we ditch
            # it and just interpret everything as `float` instead.
            # reply_list = list(map(literal_eval, reply.split(delimiter)))
            reply_list = list(map(float, reply.split(delimiter)))

        except ValueError as err:
            pft(err, 3)
            return (False, [])  # --> leaving

        except Exception as err:
            pft(err, 3)
            sys.exit(0)  # --> leaving

        return (True, reply_list)

    # --------------------------------------------------------------------------
    #   close
    # --------------------------------------------------------------------------

    def close(self, ignore_exceptions=False):
        """Cancel all pending serial operations and close the serial port."""
        if self.ser is not None:
            try:
                self.ser.cancel_read()
            except Exception:
                pass
            try:
                self.ser.cancel_write()
            except Exception:
                pass

            try:
                self.ser.close()
            except Exception as err:
                if ignore_exceptions:
                    pass
                else:
                    pft(err, 3)
                    sys.exit(0)

        self.is_alive = False

    # --------------------------------------------------------------------------
    #   connect_at_port
    # --------------------------------------------------------------------------

    def connect_at_port(self, port: str, verbose: bool = True) -> bool:
        """Open the serial port at address ``port`` and try to establish a
        connection.

        If the connection is successful and :meth:`set_ID_validation_query`
        was not set, then this function will return :const:`True`.

        If :meth:`set_ID_validation_query` was set, then this function will
        return :const:`True` or :const:`False` depending on the validation
        scheme as explained in :meth:`set_ID_validation_query`.

        Args:
            port (:obj:`str`):
                Serial port address to open.

            verbose (:obj:`bool`, optional):
                Print a `"Connecting to: "`-message to the terminal?

                Default: :const:`True`

        Returns:
            True if successful, False otherwise.
        """

        def print_success(success_str: str):
            dprint(success_str, ANSI.GREEN)
            dprint((" " * 16 + f"--> {self.name}\n"), ANSI.GREEN)

        if verbose:
            _print_hrule(True)
            if (
                self._ID_validation_query is None
                or self._valid_ID_specific is None
            ):
                msg = f"  Connecting to: {self.long_name}"
            else:
                msg = (
                    f"  Connecting to: {self.long_name} "
                    f"`{self._valid_ID_specific}`"
                )

            dprint(msg, ANSI.YELLOW)
            _print_hrule()

        print(f"  @ {port:<11s} ", end="")
        try:
            # Open the serial port
            self.ser = serial.Serial(port=port, **self.serial_settings)
        except serial.SerialException:
            print("Could not open port.")
            return False  # --> leaving
        except Exception as err:
            pft(err, 3)
            sys.exit(0)  # --> leaving

        if self._ID_validation_query is None:
            # Found any device
            print_success("Any Success!")
            self.is_alive = True
            return True  # --> leaving

        # Optional validation query
        try:
            self._force_query_to_raise_on_timeout = True
            self.is_alive = True  # We must assume communication is possible
            reply_broad, reply_specific = self._ID_validation_query()
        except Exception:
            print("Wrong or no device.")
            self.close(ignore_exceptions=True)
            return False  # --> leaving
        finally:
            self._force_query_to_raise_on_timeout = False

        if reply_broad == self._valid_ID_broad:
            if reply_specific is not None:
                print(f"Found `{reply_specific}`: ", end="")

            if self._valid_ID_specific is None:
                # Found a matching device in a broad sense
                print_success("Broad Success!")
                self.is_alive = True
                return True  # --> leaving

            if reply_specific == self._valid_ID_specific:
                # Found a matching device in a specific sense
                print_success("Specific Success!")
                self.is_alive = True
                return True  # --> leaving

        print("Wrong device.")
        self.close(ignore_exceptions=True)
        return False

    # --------------------------------------------------------------------------
    #   scan_ports
    # --------------------------------------------------------------------------

    def scan_ports(self, verbose: bool = True) -> bool:
        """Scan over all serial ports and try to establish a connection. See
        further the description at :meth:`connect_at_port`.

        Args:
            verbose (:obj:`bool`, optional):
                Print a `"Scanning ports for: "`-message to the terminal?

                Default: :const:`True`

        Returns:
            True if successful, False otherwise.
        """

        if verbose:
            _print_hrule(True)
            if (
                self._ID_validation_query is None
                or self._valid_ID_specific is None
            ):
                msg = f"  Scanning ports for: {self.long_name}"
            else:
                msg = (
                    f"  Scanning ports for: {self.long_name} "
                    f"`{self._valid_ID_specific}`"
                )
            dprint(msg, ANSI.YELLOW)
            _print_hrule()

        # Ports is a list of tuples
        ports = list(serial.tools.list_ports.comports())
        for p in ports:
            port = p[0]
            if self.connect_at_port(port, verbose=False):
                return True
            else:
                continue

        # Scanned over all the ports without finding a match
        dprint("  Error: Device not found.\n", ANSI.RED)
        return False

    # --------------------------------------------------------------------------
    #   auto_connect
    # --------------------------------------------------------------------------

    def auto_connect(
        self, filepath_last_known_port: str = "config/port.txt"
    ) -> bool:
        """Try to connect to the device using the last-known successful port
        as got written to the textfile ``filepath_last_known_port`` by the
        previous call to :meth:`auto_connect`.

        When the file does not exist, can not be read or if the desired device
        can not be found at that specific port, then a scan over all ports will
        be performed by automatically invoking :meth:`scan_ports`.

        Args:
            filepath_last_known_port (:obj:`str`):
                Default: `"config/port.txt"`

        Returns:
            True if successful, False otherwise.
        """
        path = Path(filepath_last_known_port)
        port = self._get_last_known_port(path)

        if port is None:
            if self.scan_ports():
                self._store_last_known_port(path, self.ser.portstr)
                return True

            return False

        if self.connect_at_port(port):
            self._store_last_known_port(path, self.ser.portstr)
            return True

        if self.scan_ports(verbose=False):
            self._store_last_known_port(path, self.ser.portstr)
            return True

        return False

    # -----------------------------------------------------------------------------
    #   _get_last_known_port
    # -----------------------------------------------------------------------------

    def _get_last_known_port(self, path: Path):
        """Try to open the textfile pointed to by ``path``, containing the port
        to open. Do not panic if the file does not exist or cannot be read.

        Args:
            path (:class:`pathlib.Path`):
                Path to the textfile, e.g., ``Path("config/port.txt")``.

        Returns:
            The port name :obj:`str` when the textfile is read out successfully,
            :obj:`None` otherwise.
        """
        if isinstance(path, Path):
            if path.is_file():
                try:
                    with path.open() as f:
                        port = f.readline().strip()
                    return port
                except Exception:
                    pass  # Do not panic and remain silent

        return None

    # -----------------------------------------------------------------------------
    #   _store_last_known_port
    # -----------------------------------------------------------------------------

    def _store_last_known_port(self, path: Path, port_str):
        """Try to write the port name string ``port_str`` to the textfile
        pointed to by ``path``. Do not panic if the file can not be created or
        written to.

        Args:
            path (:class:`pathlib.Path`):
                Path to the textfile, e.g., ``Path("config/port.txt")``.

            port_str (:obj:`str`):
                The port name string to write to file.

        Returns:
            True if successful, False otherwise.
        """
        if isinstance(path, Path):
            if not path.parent.is_dir():
                # Subfolder does not exists yet. Create.
                try:
                    path.parent.mkdir()
                except Exception:
                    pass  # Do not panic and remain silent

            try:
                # Write the config file
                path.write_text(port_str)
            except Exception:
                pass  # Do not panic and remain silent
            else:
                return True

        return False


def _print_hrule(leading_newline=False):
    dprint(("\n" if leading_newline else "") + "-" * 60, ANSI.YELLOW)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyQt/PySide module to provide multithreaded communication and periodical data
acquisition for a Julabo circulating bath.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=broad-except, missing-function-docstring, multiple-statements

from qtpy import QtCore, QtWidgets as QtWid
from qtpy.QtCore import Signal, Slot  # type: ignore

import dvg_pyqt_controls as controls
from dvg_debug_functions import print_fancy_traceback as pft
from dvg_qdeviceio import QDeviceIO, DAQ_TRIGGER
from dvg_devices.Julabo_circulator_protocol_RS232 import Julabo_circulator


# Enumeration
class GUI_input_fields:
    [ALL, setpoint, sub_temp, over_temp] = range(4)


class Julabo_circulator_qdev(QDeviceIO):
    """Manages multithreaded communication and periodical data acquisition for
    a Julabo circulator, referred to as the 'device'.

    In addition, it also provides PyQt/PySide GUI objects for control of the
    device. These can be incorporated into your application.

    All device I/O operations will be offloaded to 'workers', each running in
    a newly created thread.

    (*): See 'dvg_qdeviceio.QDeviceIO()' for details.

    Args:
        dev:
            Reference to a
            'dvg_devices.Julabo_circulator_protocol_RS232.Julabo_circulator'
            instance.

        debug:
            Show debug info in terminal? Warning: Slow! Do not leave on
            unintentionally.

    Main GUI objects:
        qgrp (PyQt5.QtWidgets.QGroupBox)
    """

    signal_GUI_input_field_update = Signal(int)

    def __init__(
        self,
        dev: Julabo_circulator,
        DAQ_trigger=DAQ_TRIGGER.INTERNAL_TIMER,
        DAQ_interval_ms=500,
        DAQ_timer_type=QtCore.Qt.TimerType.CoarseTimer,
        critical_not_alive_count=3,
        debug=False,
        **kwargs,
    ):
        super().__init__(dev, **kwargs)  # Pass kwargs onto QtCore.QObject()
        self.dev: Julabo_circulator  # Enforce type: removes `_NoDevice()`

        self.create_worker_DAQ(
            DAQ_trigger=DAQ_trigger,
            DAQ_function=self.DAQ_function,
            DAQ_interval_ms=DAQ_interval_ms,
            DAQ_timer_type=DAQ_timer_type,
            critical_not_alive_count=critical_not_alive_count,
            debug=debug,
        )
        self.create_worker_jobs(jobs_function=self.jobs_function, debug=debug)

        self.create_GUI()
        self.signal_DAQ_updated.connect(self.update_GUI)
        self.signal_connection_lost.connect(self.update_GUI)
        self.signal_GUI_input_field_update.connect(self.update_GUI_input_field)

        self.safe_temp.setText(f"{self.dev.state.safe_temp:.2f}")

        self.update_GUI()
        self.update_GUI_input_field()

    # --------------------------------------------------------------------------
    #   DAQ_function
    # --------------------------------------------------------------------------

    def DAQ_function(self) -> bool:
        return self.dev.query_common_readings()

    # --------------------------------------------------------------------------
    #   jobs_function
    # --------------------------------------------------------------------------

    def jobs_function(self, func, args):
        if func == "signal_GUI_input_field_update":
            # Special instruction
            self.signal_GUI_input_field_update.emit(*args)
        else:
            # Default job processing:
            # Send I/O operation to the device
            try:
                func(*args)
            except Exception as err:
                pft(err)

    # --------------------------------------------------------------------------
    #   create GUI
    # --------------------------------------------------------------------------

    def create_GUI(self):
        # Safety
        p = {
            "alignment": QtCore.Qt.AlignmentFlag.AlignRight,
            "minimumWidth": 60,
            "maximumWidth": 30,
        }
        p2 = {**p, "readOnly": True}
        self.safe_sens = QtWid.QLineEdit("nan", **p2)
        self.safe_temp = QtWid.QLineEdit("nan", **p2)
        self.sub_temp = QtWid.QLineEdit("nan", **p)
        self.sub_temp.editingFinished.connect(self.send_sub_temp_from_textbox)
        self.over_temp = QtWid.QLineEdit("nan", **p)
        self.over_temp.editingFinished.connect(self.send_over_temp_from_textbox)

        # Control
        self.pbtn_running = controls.create_Toggle_button("OFFLINE")
        self.pbtn_running.clicked.connect(self.process_pbtn_running)
        self.send_setpoint = QtWid.QLineEdit("nan", **p)
        self.send_setpoint.editingFinished.connect(
            self.send_setpoint_from_textbox
        )
        self.read_setpoint = QtWid.QLineEdit("nan", **p2)
        self.bath_temp = QtWid.QLineEdit("nan", **p2)
        self.pt100_temp = QtWid.QLineEdit("nan", **p2)
        self.status = QtWid.QPlainTextEdit()
        self.status.setReadOnly(True)
        self.status.setMaximumWidth(180)
        self.status.setStyleSheet(controls.SS_TEXTBOX_ERRORS)
        self.update_counter = QtWid.QLabel("0")

        i = 0
        p = {"alignment": QtCore.Qt.AlignmentFlag.AlignLeft}
        lbl_temp_unit = f"\u00b0{self.dev.state.temp_unit}"

        # fmt: off
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(4)

        grid.addWidget(QtWid.QLabel("<b>Safety</b>")        , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)                , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Safe sensor")          , i, 0)
        grid.addWidget(self.safe_sens                       , i, 1)
        grid.addWidget(QtWid.QLabel(lbl_temp_unit)          , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Safe temp.")           , i, 0)
        grid.addWidget(self.safe_temp                       , i, 1)
        grid.addWidget(QtWid.QLabel(lbl_temp_unit)          , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Sub temp.")            , i, 0)
        grid.addWidget(self.sub_temp                        , i, 1)
        grid.addWidget(QtWid.QLabel(lbl_temp_unit)          , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Over temp.")           , i, 0)
        grid.addWidget(self.over_temp                       , i, 1)
        grid.addWidget(QtWid.QLabel(lbl_temp_unit)          , i, 2)      ; i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)               , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("<b>Control</b>")       , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)                , i, 0)      ; i+=1
        grid.addWidget(self.pbtn_running                    , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 8)                , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Send setpoint")        , i, 0)
        grid.addWidget(self.send_setpoint                   , i, 1)
        grid.addWidget(QtWid.QLabel(lbl_temp_unit)          , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Read setpoint")        , i, 0)
        grid.addWidget(self.read_setpoint                   , i, 1)
        grid.addWidget(QtWid.QLabel(lbl_temp_unit)          , i, 2)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 8)                , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Bath temp.")           , i, 0)
        grid.addWidget(self.bath_temp                       , i, 1)
        grid.addWidget(QtWid.QLabel(lbl_temp_unit)          , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Pt100 temp.")          , i, 0)
        grid.addWidget(self.pt100_temp                      , i, 1)
        grid.addWidget(QtWid.QLabel(lbl_temp_unit)          , i, 2)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 8)                , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Status")               , i, 0, 1, 3); i+=1
        grid.addWidget(self.status                          , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 4)                , i, 0)      ; i+=1
        grid.addWidget(self.update_counter                  , i, 0, 1, 3); i+=1
        # fmt: on

        grid.setColumnStretch(0, 0)
        grid.setColumnStretch(1, 0)
        grid.setColumnStretch(2, 1)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        self.grid = grid

        self.grpb = QtWid.QGroupBox(self.dev.name)
        self.grpb.setLayout(self.grid)

    # --------------------------------------------------------------------------
    #   update_GUI
    # --------------------------------------------------------------------------

    @Slot()
    def update_GUI(self):
        """NOTE: 'self.dev.mutex' is not being locked, because we are only
        reading 'state' for displaying purposes. We can do this because 'state'
        members are written and read atomicly.
        """
        if self.dev.is_alive:
            self.pbtn_running.setChecked(bool(self.dev.state.running))
            if self.dev.state.running:
                self.pbtn_running.setText("RUNNING")
            else:
                self.pbtn_running.setText("OFF")

            self.safe_sens.setText(f"{self.dev.state.safe_sens:.2f}")
            self.read_setpoint.setText(f"{self.dev.state.setpoint:.2f}")
            self.bath_temp.setText(f"{self.dev.state.bath_temp:.2f}")
            self.pt100_temp.setText(f"{self.dev.state.pt100_temp:.2f}")

            # Check status
            self.status.setReadOnly(bool(self.dev.state.has_error))
            self.status.setStyleSheet(controls.SS_TEXTBOX_ERRORS)
            self.status.setPlainText(f"{self.dev.state.status}")

            self.update_counter.setText(f"{self.update_counter_DAQ}")
        else:
            self.pbtn_running.setText("OFFLINE")
            self.grpb.setEnabled(False)

    # --------------------------------------------------------------------------
    #   update_GUI_input_field
    # --------------------------------------------------------------------------

    @Slot()
    @Slot(int)
    def update_GUI_input_field(self, GUI_input_field=GUI_input_fields.ALL):
        if GUI_input_field == GUI_input_fields.setpoint:
            self.send_setpoint.setText(f"{self.dev.state.setpoint:.2f}")

        elif GUI_input_field == GUI_input_fields.sub_temp:
            self.sub_temp.setText(f"{self.dev.state.sub_temp:.2f}")

        elif GUI_input_field == GUI_input_fields.over_temp:
            self.over_temp.setText(f"{self.dev.state.over_temp:.2f}")

        else:
            self.send_setpoint.setText(f"{self.dev.state.setpoint:.2f}")
            self.sub_temp.setText(f"{self.dev.state.sub_temp:.2f}")
            self.over_temp.setText(f"{self.dev.state.over_temp:.2f}")

    # --------------------------------------------------------------------------
    #   GUI functions
    # --------------------------------------------------------------------------

    @Slot()
    def process_pbtn_running(self):
        if self.dev.state.running:
            self.send(self.dev.turn_off)
        else:
            self.send(self.dev.turn_on)

        # React as fast as possible by manually triggering processEvents()
        QtWid.QApplication.processEvents()

    @Slot()
    def send_setpoint_from_textbox(self):
        try:
            value = float(self.send_setpoint.text())
        except (TypeError, ValueError):
            # Revert to previously set value
            value = self.dev.state.setpoint
        except Exception as e:
            raise e

        self.add_to_jobs_queue(self.dev.set_setpoint, value)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.setpoint
        )
        self.process_jobs_queue()

    @Slot()
    def send_sub_temp_from_textbox(self):
        try:
            value = float(self.sub_temp.text())
        except (TypeError, ValueError):
            # Revert to previously set value
            value = self.dev.state.sub_temp
        except Exception as e:
            raise e

        self.add_to_jobs_queue(self.dev.set_sub_temp, value)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.sub_temp
        )
        self.process_jobs_queue()

    @Slot()
    def send_over_temp_from_textbox(self):
        try:
            value = float(self.over_temp.text())
        except (TypeError, ValueError):
            # Revert to previously set value
            value = self.dev.state.over_temp
        except Exception as e:
            raise e

        self.add_to_jobs_queue(self.dev.set_over_temp, value)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.over_temp
        )
        self.process_jobs_queue()
